###########################################################################################
#  package:   pNbody
#  file:      gh5.py
#  brief:     gh5 format (Gear in HDF5)
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>, Loic Hausammann <loic_hausammann@hotmail.com>
#
# This file is part of pNbody.
###########################################################################################


##########################################################################
#
# GEAR HDF5 CLASS
#
##########################################################################

import sys
import numpy as np

import pNbody
from pNbody import mpi, error, units
import h5py


class Nbody_arepo:

    def _init_spec(self):
        # create an empty header (in case the user is creating a dataset)
        self._header = []

    def get_excluded_extension(self):
        """
        Return a list of file to avoid when extending the default class.
        """
        return []
        
    def getParticleMatchingDict(self):
        """
        Return a list of file to avoid when extending the default class.
        """
        
        index = {
            'gas':    0,
            'halo':   1,
            'stars':  4,
            'bndry': 2}
        
        return index         

    def check_spec_ftype(self):
        try:
            if mpi.NTask > 1:
                fd = h5py.File(self.p_name_global[0],'r',driver="mpio",comm=mpi.COMM_WORLD)
            else:
                fd = h5py.File(self.p_name_global[0])

            test =         "Header"     not in fd
            test = test or "Git_commit" not in fd["Header"].attrs

            fd.close()
            
            if test:
                raise error.FormatError("arepo")

        except IOError as e:
            if self.verbose > 1:
                print("arepo not recognized: %s" % e)
            raise error.FormatError("arepo")
            
            

    def set_pio(self, pio):
        """ Overwrite function
        """
        pNbody.Nbody.set_pio(self, "no")

    def get_read_fcts(self):
        return [self.read_particles]

    def get_write_fcts(self):
        return [self.write_particles]

    def get_mxntpe(self):
        return 6

    def get_header_translation(self):
        """
        Gives a dictionnary containing all the header translation.
        If a new variable is possible in the HDF5 format, only the translation
        is required for the reader/writer.
        As h5py is not supporting dictionnary, they need special care when reading/writing.
        """
        # dict containing all the main header variables (=> easy acces)
        # e.g. self.npart will contain NumPart_ThisFile
        header_var = {}
        #header_var["Header/Code"] = "Gear"
        
        # Size variables
        header_var["Header/NumPart_ThisFile"] = "npart"
        header_var["Header/NumPart_Total"] = "nall"
        header_var["Header/NumPart_Total_HighWord"] = "nallhw"
        
        header_var["Header/MassTable"] = "massarr"
        header_var["Header/NumFilesPerSnapshot"] = "num_files"
        header_var["Header/BoxSize"] = "boxsize"
        
        header_var["Header/Git_commit"] = "git_commit"
        header_var["Header/Git_date"]   = "git_date"


        # Physics
        header_var["Header/Time"] = "atime"
        header_var["Header/Redshift"] = "redshift"
        header_var["Header/OmegaBaryon"] = "omegabaryon"
        header_var["Header/Omega0"] = "omega0"
        header_var["Header/OmegaLambda"] = "omegalambda"
        header_var["Header/HubbleParam"] = "hubbleparam"



        # Units
        header_var["Header/UnitLength_in_cm"] = "UnitLength_in_cm"
        header_var["Header/UnitMass_in_g"] = "UnitMass_in_g"
        header_var["Header/UnitVelocity_in_cm_per_s"] = "Unit_time_in_cgs"



        return header_var

    def get_list_header(self):
        """
        Gives a list of header directory from self.get_header_translation
        """
        list_header = []
        trans = self.get_header_translation()
        for key, tmp in list(trans.items()):
            directory = key.split("/")[0]
            if directory not in list_header:
                list_header.append(directory)
        return list_header

    def get_array_translation(self):
        """
        Gives a dictionnary containing all the header translation with the particles type
        requiring the array.
        If a new variable is possible in the HDF5 format, the translation,
        default value (function get_array_default_values) and dimension (function get_array_dimension)
        are required for the reader/writer.
        """
        # hdf5 names -> pNbody names
        # [name, partType, type] where partType is a list of particles with the datas
        # True means all, False means none
        # and type is the data type
        ntab = {}
        # common data
        ntab["Coordinates"]   = ["pos", True, np.float32]
        ntab["Velocities"]    = ["vel", True, np.float32]
        ntab["ParticleIDs"]   = ["num", True, np.uint32]
        ntab["Masses"]        = ["mass", True, np.float32]

        # stars data
        ntab["GFM_InitialMass"]           = ["minit", [4], np.float32]
        ntab["GFM_StellarFormationTime"]  = ["tstar", [4], np.float32]
        ntab["GFM_Metals"]                = ["metals", [4], np.float32]
        ntab["GFM_Metallicity"]           = ["metallicity", [4], np.float32]
        
        ntab["StellarHsml"]           = ["rsp", [4], np.float32]
        ntab["StellarAges"]           = ["age", [4], np.float32]
        ntab["MH"]                    = ["mh",  [4], np.float32]

        
        # magnitues
        ntab["MagF475X"]              = ["MagF475X",    [4], np.float32]
        ntab["MagVISeuclid"]          = ["MagVISeuclid",[4], np.float32]
        ntab["MagYeuclid"]            = ["MagYeuclid",  [4], np.float32]
        ntab["MagJeuclid"]            = ["MagJeuclid",  [4], np.float32]
        
        return ntab


    def get_array_default_value(self):
        """
        Gives a dictionary of default value for pNbody's arrays
        """
        # default value
        dval = {}
        dval["pos"] = 0.0
        dval["vel"] = 0.0
        dval["num"] = 0.0
        dval["mass"] = 0.0
        dval["u"] = 0.0
        dval["rho"] = 0.0
        dval["metals"] = 0.0
        dval["metallicity"] = 0.0
        dval["opt1"] = 0.0
        dval["opt2"] = 0.0
        dval["rsp"] = 0.0
        dval["minit"] = 0.0
        dval["tstar"] = 0.0
        dval["idp"] = 0.0
        dval["rsp_stars"] = 0.0
        dval["metals_stars"] = 0.0
        dval["metallicity_stars"] = 0.0
        dval["rho_stars"] = 0.0
        dval["snii_thermal_time"] = 0.0
        dval["snia_thermal_time"] = 0.0
        dval["xdi"] = 0.
        dval["xdii"] = 0.
        dval["xe"] = 0.
        dval["xh2i"] = 0.
        dval["xh2ii"] = 0.
        dval["xhdi"] = 0.
        dval["xhi"] = 0.
        dval["xhii"] = 0.
        dval["xhm"] = 0.
        dval["xhei"] = 0.
        dval["xheii"] = 0.
        dval["xheiii"] = 0.
        dval["Lr"] = 0.0
        dval["Lv"] = 0.0
        dval["Lb"] = 0.0
        return dval

    def get_array_dimension(self):
        """
        Gives a dictionary of dimension for pNbody's arrays
        """
        # dimension
        vdim = {}
        vdim["pos"] = 3
        vdim["vel"] = 3

        vdim["num"] = 1
        vdim["mass"] = 1
        vdim["u"] = 1
        vdim["rho"] = 1
        vdim["opt1"] = 1
        vdim["opt2"] = 1
        vdim["rsp"] = 1
        vdim["minit"] = 1
        vdim["tstar"] = 1
        vdim["idp"] = 1
        vdim["rsp_stars"] = 1
        vdim["rho_stars"] = 1
        vdim["snii_thermal_time"] = 1
        vdim["snia_thermal_time"] = 1
        vdim["xdi"] = 1
        vdim["xdii"] = 1
        vdim["xe"] = 1
        vdim["xh2i"] = 1
        vdim["xh2ii"] = 1
        vdim["xhdi"] = 1
        vdim["xhi"] = 1
        vdim["xhii"] = 1
        vdim["xhm"] = 1
        vdim["xhei"] = 1
        vdim["xheii"] = 1
        vdim["xheiii"] = 1

        vdim["metals"] = self.ChimieNelements
        vdim["metals_stars"] = self.ChimieNelements
        vdim["metallicity"] = 1
        vdim["Lr"] = 1
        vdim["Lv"] = 1
        vdim["Lb"] = 1        
        return vdim

    def get_default_spec_vars(self):
        '''
        return specific variables default values for the class
        '''

        return {'massarr': np.array([0, 0, self.nbody, 0, 0, 0]),
                'atime': 0.,
                'redshift': 0.,
                'flag_sfr': 0,
                'flag_feedback': 0,
                'nall': np.array([0, 0, self.nbody, 0, 0, 0]),
                'npart': np.array([0, 0, self.nbody, 0, 0, 0]),
                'flag_cooling': 0,
                'num_files': 1,
                'boxsize': 0.,
                'omega0': 0.,
                'omegalambda': 0.,
                'hubbleparam': 0.,
                'flag_age': 0.,
                'hubbleparam': 0.,
                'flag_metals': 0.,
                'nallhw': np.array([0, 0, 0, 0, 0, 0]),
                'flag_entr_ic': 0,
                'flag_chimie_extraheader': 0,
                'critical_energy_spec': 0.,
                'flag_thermaltime': 0,
                'empty': '',
                'comovingintegration': True,
                'hubblefactorcorrection': True,
                'comovingtoproperconversion': True,
                'ChimieNelements': 0,
                'utype':"gear",
                }

    def get_particles_limits(self, i):
        """ Gives the limits for a thread.
        In order to get the particles, slice them like this pos[start:end].
        :param int i: Particle type
        :returns: (start, end)
        """
        nber = float(self.nall[i]) / mpi.mpi_NTask()
        start = int(mpi.mpi_ThisTask() * nber)
        end = int((mpi.mpi_ThisTask() + 1) * nber)
        return start, end

    def set_local_value(self):
        N = mpi.mpi_NTask()
        if N == 1:
            return
        part = len(self.nall)
        for i in range(part):
            s, e = self.get_particles_limits(i)
            self.npart[i] = e - s

    def get_massarr_and_nzero(self):
        """
        return massarr and nzero

        !!! when used in //, if only a proc has a star particle,
        !!! nzero is set to 1 for all cpu, while massarr has a length of zero !!!
        """

        if self.has_var('massarr') and self.has_var('nzero'):
            if self.massarr is not None and self.nzero is not None:
                self.warning("warning : get_massarr_and_nzero : here we use massarr and nzero %s %s"%(str(self.massarr),str(self.nzero)),verbosity=2)
                return self.massarr, self.nzero

        massarr = zeros(len(self.npart), float)
        nzero = 0

        # for each particle type, see if masses are equal
        for i in range(len(self.npart)):
            first_elt = sum((arange(len(self.npart)) < i) * self.npart)
            last_elt = first_elt + self.npart[i]

            if first_elt != last_elt:
                c = (self.mass[first_elt] ==
                     self.mass[first_elt:last_elt]).astype(int)
                if sum(c) == len(c):
                    massarr[i] = self.mass[first_elt]
                else:
                    nzero = nzero + len(c)

        return massarr.tolist(), nzero

    def read_particles(self, f):
        '''
        read gadget file
        '''
        from copy import deepcopy

        # go to the end of the file
        if f is not None:
            f.seek(0, 2)

        import h5py
        if mpi.mpi_NTask() > 1:
            fd = h5py.File(
                self.p_name_global[0],
                'r',
                driver="mpio",
                comm=mpi.comm)
        else:
            fd = h5py.File(self.p_name_global[0], 'r')

        ################
        # read header
        ################
        if self.verbose > 0 and mpi.mpi_IsMaster():
            print("reading header...")

        # set default values
        default = self.get_default_spec_vars()
        for key, i in list(default.items()):
            setattr(self, key, i)

        # get values from snapshot
        trans = self.get_header_translation()

        list_header = self.get_list_header()

        for name in list_header:
            # e.g. create self.npart with the value
            # fd["Header"].attrs["NumPart_ThisFile"]
            for key in fd[name].attrs:
                full_name = name + "/" + key
                if full_name not in list(trans.keys()):
                    print((
                        "'%s' key not recognized, please add it to config/format/gh5.py file in get_header_translation" %
                        full_name))
                    trans[full_name] = key
                tmp = fd[name].attrs[key]
                if isinstance(tmp, bytes) and tmp == "None":
                    tmp = None
                if isinstance(tmp, bytes):
                    tmp = tmp.decode('utf-8', "ignore")
                setattr(self, trans[full_name], tmp)


        # elements
          
        self.ChimieNelements = 10
        self.ChimieElements = ['H', 'He', 'C', 'N', 'O', 'Ne', 'Mg', 'Si', 'Fe','Untracked']
        self.ChimieSolarMassAbundances = {}
        self.ChimieSolarMassAbundances['Fe'] = 0.001771
        self.ChimieSolarMassAbundances['Mg'] = 0.00091245
        self.ChimieSolarMassAbundances['O'] = 0.0108169
        self.ChimieSolarMassAbundances['Metals'] = 0.0127  # value used by arepo

        # get local value from global ones
        self.set_local_value()

        ###############
        # read units
        ###############
        if self.verbose > 0 and mpi.mpi_IsMaster():
            print("reading units...")

        # define system of units
        params = {}

        # consider that if we have length unit, we should have them all
        if hasattr(self, "UnitLength_in_cm"):
            params['UnitLength_in_cm'] = float(self.UnitLength_in_cm)
            if hasattr(self, "UnitVelocity_in_cm_per_s"):
                params['UnitVelocity_in_cm_per_s'] = float(
                    self.UnitVelocity_in_cm_per_s)
            else:
                self.UnitVelocity_in_cm_per_s = float(
                    self.UnitLength_in_cm) / float(self.Unit_time_in_cgs)
                params['UnitVelocity_in_cm_per_s'] = self.UnitVelocity_in_cm_per_s
            params['UnitMass_in_g'] = float(self.UnitMass_in_g)
            self.localsystem_of_units = units.Set_SystemUnits_From_Params(
                params)
        else:
            print("WARNING: snapshot seems broken! There is no units!")

        ################
        # read particles
        ################

        if self.verbose > 0 and mpi.mpi_IsMaster():
            print("reading particles...")

        def getsize(n, dim):
            if dim == 1:
                return (n,)
            if dim > 1:
                return (n, dim)

        list_of_vectors = []

        # check what vector we needs to be created

        ntab = self.get_array_translation()
        vdim = self.get_array_dimension()
        dval = self.get_array_default_value()

        for i, n in enumerate(self.npart):

            if n == 0:
                continue

            if list(fd.keys()).count("PartType%d" % i) == 0:
                raise IOError("type=%d n=%d but group %s is not found !" %
                              (i, n, "PartType%d" % i))

            # loop over dataset
            block = fd["PartType%d" % i]
            for key in list(block.keys()):
                if key not in ntab:
                    print("get a dataset with name %s but no such key in ntab" %key)
                    
                    continue

                varname = ntab[key][0]

                if self.has_var(varname):
                    pass  # do nothing
                else:
                    setattr(self, varname, None)

                # record the variable in a list
                if list_of_vectors.count(varname) == 0:
                    list_of_vectors.append(varname)

        # read the tables
        for i, n in enumerate(self.npart):

            if n == 0:
                continue

            if self.verbose > 1 and mpi.mpi_IsMaster():
                print(("Reading particles (type %i)..." % i))

            remaining_vectors = deepcopy(list_of_vectors)

            # loop over dataset
            block = fd["PartType%d" % i]
            if mpi.mpi_NTask() > 1:
                init, end = self.get_particles_limits(i)

            for key in list(block.keys()):
                if key not in ntab:
                    continue
                varname = ntab[key][0]
                if self.skip_io_block(varname):
                    if self.verbose > 1:
                        print(("Skipping %s block" % varname))
                    continue
                else:
                    if self.verbose > 1:
                        print("Reading %s block" % varname)
                var = getattr(self, varname)

                if mpi.mpi_NTask() > 1:
                    data = block[key][init:end]
                else:
                    data = block[key][()]
                if var is not None:
                    setattr(self, varname, np.concatenate(
                        (getattr(self, varname), data)))
                else:
                    setattr(self, varname, data)

                remaining_vectors.remove(varname)

            # loop over vectors absent from the block
            for varname in remaining_vectors:
                var = getattr(self, varname)
                data = (
                    np.ones(getsize(n, vdim[varname])) *
                    dval[varname]).astype(np.float32)
                if var is not None:
                    setattr(self, varname, np.concatenate(
                        (getattr(self, varname), data)))
                else:
                    setattr(self, varname, data)

        # set tpe
        self.tpe = np.array([], np.int32)
        for i in range(len(self.npart)):
            self.tpe = np.concatenate((self.tpe, np.ones(self.npart[i]) * i))

        # compute nzero
        nzero = 0
        mass = np.array([])

        for i in range(len(self.npart)):
            if self.massarr is None or self.massarr[i] == 0:
                nzero = nzero + self.npart[i]
            else:
              self.warning("Massarr is not supported! Please specify the mass of all the particles!",verbosity=2)

        self.nzero = nzero

        mpi.mpi_barrier()
        fd.close()
        
        
        # additional initialization
        if self.comovingintegration:
          self.comovingintegration=True
          self.hubblefactorcorrection=True
        else:
          self.comovingintegration=False
          self.hubblefactorcorrection=False
          
        if self.comovingintegration:
          self.comovingtoproperconversion  = True 
        
        
        

    def write_particles(self, f):
        '''
        specific format for particle file
        '''
        # go to the end of the file
        if f is not None:
            f.close()

        # not clean, but work around pNbody
        filename = self.p_name_global[0]
        # open file
        if mpi.mpi_NTask() > 1:
            h5f = h5py.File(filename, "w", driver="mpio", comm=mpi.comm)
        else:
            h5f = h5py.File(filename, "w")

        # add units to the usual gh5 struct
        if hasattr(self, "unitsparameters"):
            units = self.unitsparameters.get_dic()
            for key, i in list(units.items()):
                if not hasattr(self, key):
                    setattr(self, key, i)

        if hasattr(self, "UnitVelocity_in_cm_per_s") and hasattr(
                self, "UnitLength_in_cm"):
            self.Unit_time_in_cgs = self.UnitLength_in_cm / self.UnitVelocity_in_cm_per_s

        ############
        # HEADER
        ############
        if self.verbose > 0 and mpi.mpi_IsMaster():
            print("Writing header...")

        if self.massarr is None:
            self.massarr = np.zeros(len(self.npart))

        list_header = self.get_list_header()
        trans = self.get_header_translation()
        
        
        # cheat a little bit in order to get the real number of particles in
        # this file
        trans["Header/NumPart_ThisFile"] = "npart_tot"

        for name in list_header:
            h5f.create_group(name)
        for key in self.get_list_of_vars():
            if key in list(trans.values()):
                ind = list(trans.values()).index(key)
                name, hdf5 = list(trans.keys())[ind].split("/")
                value = getattr(self, key)
                if not isinstance(value, dict):
                    if value is None:
                        h5f[name].attrs[hdf5] = "None"
                    else:
                        print(name, hdf5, value)
                        if hdf5 == "ChimieLabels":
                            value = ','.join(value)
                        h5f[name].attrs[hdf5] = value


        ##############
        # PARTICULES
        ##############
        if self.verbose > 0 and mpi.mpi_IsMaster():
            print("Writing particles...")

        for i in range(len(self.npart)):
            if self.massarr is not None and self.massarr[i] != 0:
                self.warning("Massarr is not supported! Please specify the mass of all the particles!",verbosity=2)

        ntab = self.get_array_translation()
        # get particles type present
        type_part = []
        for i in range(len(self.npart)):
            if self.npart[i] > 0:
                type_part.append(i)

        # write particles
        for i in type_part:
            if mpi.mpi_NTask() > 1:
                init, end = self.get_particles_limits(i)
            if self.verbose > 1 and mpi.mpi_IsMaster():
                print("Writing particles (type %i)..." % i)
            group = "PartType%i" % i
            grp = h5f.create_group(group)
            nb_sel = self.select(i)
            for key, j in list(ntab.items()):
                            
                varname = j[0]
                var_type = j[1]
                
                if (var_type is True) or i in var_type:
                    if hasattr(nb_sel, varname) and getattr(nb_sel, varname) is not None:
                        # get and transform type
                        tmp = getattr(nb_sel, varname)
                        tmp = tmp.astype(j[2])
                        # create dataset
                        size = list(tmp.shape)
                        size[0] = self.npart_tot[i]
                        if mpi.mpi_NTask() > 1:
                            dset = grp.create_dataset(key, size, dtype=j[2])
                            dset[init:end] = tmp
                        else:
                            h5f[group][key] = tmp

        h5f.close()



###########################################################################################
#  package:   Mockimgs
#  file:      filters.py
#  brief:     filter class
#  copyright: GPLv3
#             Copyright (C) 2023 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

import sys
import os
import numpy as np

from astropy import units as u
from astropy import constants as c
from scipy import integrate

from ..parameters import FILTERSDIR
from ..interpolator import SSPInterpolator
from ..iofunc import SSPGrid 
from .spectra import Spectra 

filters_dic = {}

# magnitude
filters_dic['GAEA_B']          = (os.path.join(FILTERSDIR,"ages_mag_B.dat"),  "GAEA")
filters_dic['GAEA_VIS']        = (os.path.join(FILTERSDIR,"ages_mag_VIS.dat"),"GAEA")
filters_dic['GAEA_Y']          = (os.path.join(FILTERSDIR,"ages_mag_Y.dat"),  "GAEA")
filters_dic['GAEA_J']          = (os.path.join(FILTERSDIR,"ages_mag_J.dat"),  "GAEA")

filters_dic['CMD_F475X']       = (os.path.join(FILTERSDIR,"HSTF475Xmag_CMD_1e6_patched.pkl"),   "CMD")
filters_dic['CMD_VIS']         = (os.path.join(FILTERSDIR,"EuclidVISmag_CMD_1e6_patched.pkl"),  "CMD")
filters_dic['CMD_VISb']        = (os.path.join(FILTERSDIR,"EuclidVISmag_CMD_1e6_b_patched.pkl"),"CMD")
filters_dic['CMD_Y']           = (os.path.join(FILTERSDIR,"EuclidYmag_CMD_1e6_patched.pkl"),    "CMD")
filters_dic['CMD_J']           = (os.path.join(FILTERSDIR,"EuclidJmag_CMD_1e6_patched.pkl"),    "CMD")

filters_dic['BastI_GAIA_G']    = (os.path.join(FILTERSDIR,"GAIA_G_BastI_1e6.hdf5"),    "hdf5")
filters_dic['BastI_GAIA_G_RP'] = (os.path.join(FILTERSDIR,"GAIA_G_RP_BastI_1e6.hdf5"), "hdf5")
filters_dic['BastI_GAIA_G_BP'] = (os.path.join(FILTERSDIR,"GAIA_G_BP_BastI_1e6.hdf5"), "hdf5")
filters_dic['BastI_SDSS_u']    = (os.path.join(FILTERSDIR,"SDSS_u_BastI_1e6.hdf5"),    "hdf5")
filters_dic['BastI_SDSS_g']    = (os.path.join(FILTERSDIR,"SDSS_g_BastI_1e6.hdf5"),    "hdf5")
filters_dic['BastI_SDSS_r']    = (os.path.join(FILTERSDIR,"SDSS_r_BastI_1e6.hdf5"),    "hdf5")
filters_dic['BastI_SDSS_i']    = (os.path.join(FILTERSDIR,"SDSS_i_BastI_1e6.hdf5"),    "hdf5")
filters_dic['BastI_SDSS_z']    = (os.path.join(FILTERSDIR,"SDSS_z_BastI_1e6.hdf5"),    "hdf5")
filters_dic['BastI_HST_F475X'] = (os.path.join(FILTERSDIR,"HST_F475X_BastI_1e6.hdf5"), "hdf5")
filters_dic['BastI_Euclid_VIS']= (os.path.join(FILTERSDIR,"Euclid_VIS_BastI_1e6.hdf5"),"hdf5")
filters_dic['BastI_Euclid_J']  = (os.path.join(FILTERSDIR,"Euclid_J_BastI_1e6.hdf5"),  "hdf5")
filters_dic['BastI_Euclid_Y']  = (os.path.join(FILTERSDIR,"Euclid_Y_BastI_1e6.hdf5"),  "hdf5")
filters_dic['BastI_Euclid_H']  = (os.path.join(FILTERSDIR,"Euclid_H_BastI_1e6.hdf5"),  "hdf5")
filters_dic['BastI_Jonhson_V']  = (os.path.join(FILTERSDIR,"Johnson_V_BastI_1e6.hdf5"),  "hdf5")


filters_dic['SB99_SDSS_u']    = (os.path.join(FILTERSDIR,"SB99_SDSSu.hdf5"),    "hdf5")
filters_dic['SB99_SDSS_g']    = (os.path.join(FILTERSDIR,"SB99_SDSSg.hdf5"),    "hdf5")
filters_dic['SB99_SDSS_r']    = (os.path.join(FILTERSDIR,"SB99_SDSSr.hdf5"),    "hdf5")
filters_dic['SB99_SDSS_i']    = (os.path.join(FILTERSDIR,"SB99_SDSSi.hdf5"),    "hdf5")
filters_dic['SB99_SDSS_z']    = (os.path.join(FILTERSDIR,"SB99_SDSSz.hdf5"),    "hdf5")

filters_dic['SB99_ARK_VIS1']    = (os.path.join(FILTERSDIR,"SB99_ARKVIS1.hdf5"),    "hdf5")
filters_dic['SB99_ARK_VIS2']    = (os.path.join(FILTERSDIR,"SB99_ARKVIS2.hdf5"),    "hdf5")
filters_dic['SB99_ARK_NIR1']    = (os.path.join(FILTERSDIR,"SB99_ARKNIR1.hdf5"),    "hdf5")
filters_dic['SB99_ARK_NIR2']    = (os.path.join(FILTERSDIR,"SB99_ARKNIR2.hdf5"),    "hdf5")

filters_dic['BPASS230_ARK_VIS1']    = (os.path.join(FILTERSDIR,"BPASS230_ARKVIS1.hdf5"),    "hdf5")
filters_dic['BPASS230_ARK_VIS2']    = (os.path.join(FILTERSDIR,"BPASS230_ARKVIS2.hdf5"),    "hdf5")
filters_dic['BPASS230_ARK_VIS1bc']    = (os.path.join(FILTERSDIR,"BPASS230_ARKVIS1bc.hdf5"),    "hdf5")
filters_dic['BPASS230_ARK_VIS2bc']    = (os.path.join(FILTERSDIR,"BPASS230_ARKVIS2bc.hdf5"),    "hdf5")
filters_dic['BPASS230_ARK_NIR1']    = (os.path.join(FILTERSDIR,"BPASS230_ARKNIR1.hdf5"),    "hdf5")
filters_dic['BPASS230_ARK_NIR2']    = (os.path.join(FILTERSDIR,"BPASS230_ARKNIR2.hdf5"),    "hdf5")


filters_dic['BPASS221_SDSS_u']    = (os.path.join(FILTERSDIR,"BPASS221_SDSSu.hdf5"),    "hdf5")
filters_dic['BPASS221_SDSS_g']    = (os.path.join(FILTERSDIR,"BPASS221_SDSSg.hdf5"),    "hdf5")
filters_dic['BPASS221_SDSS_r']    = (os.path.join(FILTERSDIR,"BPASS221_SDSSr.hdf5"),    "hdf5")
filters_dic['BPASS221_SDSS_i']    = (os.path.join(FILTERSDIR,"BPASS221_SDSSi.hdf5"),    "hdf5")
filters_dic['BPASS221_SDSS_z']    = (os.path.join(FILTERSDIR,"BPASS221_SDSSz.hdf5"),    "hdf5")

filters_dic['BPASS230_SDSS_u']    = (os.path.join(FILTERSDIR,"BPASS230_SDSSu.hdf5"),    "hdf5")
filters_dic['BPASS230_SDSS_g']    = (os.path.join(FILTERSDIR,"BPASS230_SDSSg.hdf5"),    "hdf5")
filters_dic['BPASS230_SDSS_r']    = (os.path.join(FILTERSDIR,"BPASS230_SDSSr.hdf5"),    "hdf5")
filters_dic['BPASS230_SDSS_i']    = (os.path.join(FILTERSDIR,"BPASS230_SDSSi.hdf5"),    "hdf5")
filters_dic['BPASS230_SDSS_z']    = (os.path.join(FILTERSDIR,"BPASS230_SDSSz.hdf5"),    "hdf5")

filters_dic['BPASS230_Euclid_Y']    = (os.path.join(FILTERSDIR,"BPASS230_EuclidNISPY.hdf5"),    "hdf5")
filters_dic['BPASS230_Euclid_J']    = (os.path.join(FILTERSDIR,"BPASS230_EuclidNISPJ.hdf5"),    "hdf5")
filters_dic['BPASS230_Euclid_H']    = (os.path.join(FILTERSDIR,"BPASS230_EuclidNISPH.hdf5"),    "hdf5")
filters_dic['BPASS230_Euclid_VIS']  = (os.path.join(FILTERSDIR,"BPASS230_EuclidVIS.hdf5"),      "hdf5")

filters_dic['BPASS230_JKC_U']  = (os.path.join(FILTERSDIR,"BPASS230_JKC_U.hdf5"),      "hdf5")
filters_dic['BPASS230_JKC_V']  = (os.path.join(FILTERSDIR,"BPASS230_JKC_V.hdf5"),      "hdf5")



default = "BastI_SDSS_g"


def List():
  """
  return a list of current available filters
  """  
  for name in filters_dic.keys():
    print(name)
    
    
def list():
  """
  return a list of current available filters
  """
  for k in filters_dic.keys():
    print(k)
          

class FilterCl(Spectra):
  '''
  Define a filter class that describe a filter from its transmission curve.
  
  filters are characterised by their transmission curve given as a function
  of the wavelength (by default in angstrom)
  
  '''

  def __init__(self,filter_filename):
    
    self.filter_filename = filter_filename
    self.read()    
    self.init()
    

  def read(self):
    """
    Read the file that contains the filter information
    and return the filter transmission
    as a function of the wavelength in Angstrom
    """
    data = np.loadtxt(self.filter_filename)
    # wavelength in Agstrom
    self.wavelength = data[:,0]*u.angstrom
    self.data       = data[:,1]
  
  
  def init(self):
    """
    init a set of useful quantities defining the properties of the filter
    """
    
    l = self.wavelength.to(u.angstrom).value
    R = self.data
    
    # Lambda mean (0.4719 micrometer)
    I1 = integrate.simpson(l*R,l)
    I2 = integrate.simpson(R,l)
    self.wavelength_mean = (I1/I2 * u.angstrom).to(u.micrometer)
    #print("lambda_mean [micrometer]",lambda_mean)
    
    
    # Lambda pivot (0.4702 micrometer)
    I1 = integrate.simpson(l*R,l)
    I2 = integrate.simpson(R/l,l)
    self.wavelength_pivot = (np.sqrt(I1/I2) * u.angstrom).to(u.micrometer)
    #print("lambda_pivot [micrometer]",lambda_pivot)
    
    # Lambda n1 (0.4751 micrometer)
    I1 = integrate.simpson(R*l**2,l)
    I2 = integrate.simpson(R*l   ,l)
    self.wavelength_n1 = (I1/I2 * u.angstrom).to(u.micrometer)
    #print("lambda_n1 [micrometer]",lambda_n1)
    
    # Lambda n2 (0.4686 micrometer)
    I1 = integrate.simpson(R,   l)
    I2 = integrate.simpson(R/l, l)
    self.wavelength_n2 = (I1/I2 * u.angstrom).to(u.micrometer)
    #print("lambda_n2 [micrometer]",lambda_n2)  
    
    # delta Lambda
    I1 = integrate.simpson(R,   l)
    self.delta_wavelength = I1/np.max(R) * u.angstrom

    # lambda_min
    self.wavelength_min = (self.wavelength_pivot - self.delta_wavelength/2).to(u.angstrom)
    #self.wavelength_min = min(l) * u.angstrom
    
    # lambda_max
    self.wavelength_max = (self.wavelength_pivot + self.delta_wavelength/2).to(u.angstrom)
    #self.wavelength_max = max(l) * u.angstrom
    
    # response maxW
    self.response_max = np.max(R)
    
    # total response
    c = (l>=self.wavelength_min.value)*(l<=self.wavelength_max.value)
    lc = np.compress(c,l)
    Rc = np.compress(c,R)
    I1 = integrate.simpson(Rc,   lc)
    I2 = lc[-1] - lc[0]
    self.TotalResponse = I1/I2
    
  
  def SEDmultiply(self,SED,wavelength):
    """
    multipy a given SED with the current filter
    
    SED        : the SED in arbitrary units
    wavelength : wavelength in angstrom 
    """
    
  
  def get_wavelength(self,units=u.angstrom):
    """
    get the wavelength
    """
    return self.wavelength.to(units)
  
  def get_response(self):
    """
    get the filter response
    """
    return self.get_data(units=None)

  


  
    

def Filter(name,filter_type=None):
  """
  the filter class builder
  """
  
  if os.path.isfile(name):
    filename = name
    if filter_type is None:
      # try to find the filter type
      ext=os.path.splitext(name)[1]
      if   ext==".HDF5" or ext==".hdf5":
        filter_type="hdf5"
      elif ext==".PKL" or  ext==".pkl":
        filter_type="CMD"
      else:
        raise NameError("Unrecognized filter type %s"%name) 
  else:
    filename,filter_type = filters_dic[name]  
  
  if   filter_type == "GAEA":
    return FilterGAEA(filename,filter_name=name)
  
  elif filter_type == "CMD" or filter_type == "BastI" or filter_type == "PKL":
    return FilterPKL(filename,filter_name=name)

  elif filter_type == "hdf5":
    return FilterHDF5(filename,filter_name=name)
           
  raise NameError("Unknown filter %s"%name) 



class FilterClass():
  """
  the filter class
  """
  
  def __init__(self,filename,filter_name=None):
    """
    filtername : name of the filter, e.g. GAEA_B, GAEA_VIS, BastI_GAIA_G
    """
    
    self.file = filename
    self.name = filter_name
    
    # read the file containing the filter information
    Ages,Zs,Mat,MfinMini = self.Read()
    
    # make global
    self.Mat = Mat
    
    # create the interpolator
    self.interpolator = SSPInterpolator(Mat,Ages,Zs)
    
    # create the Mass ratio interpolator
    if MfinMini is not None:
      self.Mratio_interpolator = SSPInterpolator(MfinMini,Ages,Zs)
    


  def info(self):
    print("filter class          : %s"%self.__class__.__name__)
    print("filter name           : %s"%self.name)
    
  def get_parameters(self,fmt=None):
    """
    return a dictionary containing useful parameters    
    """
    params = {}
    if fmt=="fits":
      params["FILTER"]           = (self.name,"Filter name")
    else:  
      params["filter_name"]      = self.name
    
    return params

  def Read(self):
    """
    read the file and create a table
    """
    pass




class FilterGAEA(FilterClass):
  """
  the filter GAEA class
  """
  
  def Read(self):
    """
    special read for FilterGAEA
    """
    from astropy.io import ascii 
    data = ascii.read(self.file)

    # from data extract metalicites (zs) ages (ages) and ML (vs)
    key_zs = [22, 32, 42, 52, 62, 72] # see http://www.bruzual.org/bc03/doc/bc03.pdf
    zs = [-2.2490, -1.6464, -0.6392, -0.3300, 0.0932, 0.5595]
    ages  = []
    for i in range(0,220):
      ages.append(data[i][0])

    Mat = np.zeros((len(ages), len(zs)))

    for i in range(len(ages)):
      for j in range(len(zs)):
        Mat[i][j] = data[i][j+1]

    Mat = Mat.transpose() # transpose to agree with the convention in libvazdekis.py
    
    # here Mat contains magnitudes
    return ages, zs, Mat, None


  def Magnitudes(self,mass, age, Z):
    '''
    Ages are provided in Gyrs 
    Mass in Msol
    '''  
    # convert ages to yrs
    age = age*1e9
    
    # get magnitudes
    M = self.interpolator.Get(age,Z)
    # normalize (tables assume 1e11 Msol)
    F = 10**(-M*0.4)*mass/1e11
    # back to magnitudes
    Mag = -2.5*np.log10(F)
    return Mag



class FilterPKL(FilterClass):
  """
  the Filter PKL class
  """
  
  def Read(self):
    """
    special read for FilterPKL
    """
    import pickle
  
    # read the pickle file  
    f = open(self.file,"rb")
    data = pickle.load(f)
    binsAge = pickle.load(f)
    binsFe  = pickle.load(f)
    f.close()

    Mat = np.transpose(data)    
    Ages = binsAge
    Zs   = binsFe   
    
    return Ages,Zs, Mat, None


  def Magnitudes(self,mass, age, Z):
    '''
    Ages are provided in Gyrs 
    '''  
    # convert ages to Myrs
    age = age*1e3
        
    # get magnitudes
    M = self.interpolator.Get(age,Z)
    # transform to a flux
    F  = 10**(-M/2.5)
    F = F*mass
    # back to magnitudes
    Mag = -2.5*np.log10(F)
    
    return Mag



class FilterHDF5(FilterClass):
  """
  the Filter HDF5 class
  """
  
  def Read(self):
    """
    special read for FilterPKL
    """
    SSP = SSPGrid(self.file)
    data = SSP.read()

    Mat       = np.transpose(data["Data"]["Magnitudes"])
    MfinMini  = np.transpose(data["Data"]["MiniMfin"])    
    Ages      = np.transpose(data["Data"]["Ages"])
    Zs        = np.transpose(data["Data"]["MH"])   
    
    return Ages,Zs,Mat,MfinMini


  def Magnitudes(self,mass, age, MH, current_mass=False):
    '''
    mass       : initial SSP mass in Msol
    age        : SSP age in Gyr
    MH         : SSP metallicity in log10(M/M0)

    current_mass : if True, assume the mass to be the final SSP mass
                   and convert it to the initial SSP mass
                   default=False
                 : if False, assume the mass is the initial SSP mass
                 
    '''  
    # convert ages to Myrs
    age = age*1e3
    
    # convert final mass to initial IMF mass
    if current_mass is True:
      mass = mass *self.Mratio_interpolator.Get(age,MH)  
        
    # get magnitudes
    M = self.interpolator.Get(age,MH)
    # transform to a flux
    F  = 10**(-M/2.5)
    F = F*mass
    # back to magnitudes
    Mag = -2.5*np.log10(F)
        
    return Mag


  def SSPMassRatio(self, age, MH):
    """
    Return the ratio of the initial SSP to the final SSP mass
    """
    # convert ages to Myrs
    age = age*1e3
    return self.Mratio_interpolator.Get(age,MH)





'''
 @package   pNbody
 @file      setup.py
 @brief     Install pNbody
 @copyright GPLv3
 @author    Yves Revaz <yves.revaz@epfl.ch>
 @section   COPYRIGHT  Copyright (C) 2017 EPFL (Ecole Polytechnique Federale de Lausanne)  LASTRO - Laboratory of Astrophysics of EPFL

 This file is part of pNbody.
'''

import os
import sys
from setuptools import setup, find_packages, Extension
from glob import glob
from copy import deepcopy
import versioneer
import numpy as np


# get and save the current git version
#versioneer.get_version("pNbody")


# optional modules
options = {
    "ptreelib": False,
    "cooling_with_metals": True,
    "pygsl": True,
}

# special flags for modules
spec_flags = {
    # "logger_loader": ["-Wall", "-Wextra", "-DSWIFT_DEBUG_CHECKS"]
}

# default flags
flags = ["-Werror=strict-prototypes", "-Wno-unused-parameter"]

# directory containing headers
include_dirs = [
    ".",
    "/usr/include",     # gsl
    "/usr/local/include",     # macOS (homebrew) headers
    np.get_include(),
]

# special includes (not used)
spec_inc = {
    "ptreelib": ["$MPICH_ROOT/include",
                 "src/ptreelib"],

}

# libraries
libraries = ["m"]

# special libraries
spec_lib = {
    "ptreelib": ["mpich"],
    "cooling_with_metals": ["gsl", "gslcblas"],
    "pygsl": ["gsl"]
}

# special libraries directory
spec_lib_dir = {
    "ptreelib": ["$MPICH_ROOT/lib"]
}

# C modules
modules_name = [
    "nbodymodule",
    "myNumeric",
    "mapping",
    "kernels",
    "montecarlolib",
    "iclib",
    "treelib",
    "nbdrklib",
    "peanolib",
    "coolinglib",
    "cosmolib",
    "asciilib",
    "tessel",
    "pychem",
    # "libtipsy",    crash which fedora 23: needs CFLAGS="-I/usr/include/tirpc" CXXFLAGS="-I/usr/include/tirpc"
    "thermodynlib",
    "orbitslib"
]

# data files
'''
data_files = []

data_files.append(('config', glob('./config/*parameters')))
data_files.append(('config/rgb_tables', glob('./config/rgb_tables/*')))
# # trick to avoid rpm problems
data_files.append(('config/formats', glob('./config/formats/*.py')))
data_files.append(('config/extensions', glob('./config/extensions/*.py')))

data_files.append(('plugins', glob('./config/*.py')))
data_files.append(('plugins', glob('./config/plugins/*.py')))
data_files.append(('config/opt/SSP', glob('./config/opt/SSP/*[txt,dat]')))
data_files.append(('config/opt/SSP/P94_salpeter',
                   glob('./config/opt/SSP/P94_salpeter/*')))
data_files.append(('fonts', glob('./fonts/*')))

# examples
data_files.append(('examples', glob('./examples/*.dat')))
data_files.append(('examples', glob('./examples/*.py')))
data_files.append(('examples/scripts', glob('./examples/scripts/*.py')))

# examples ic
data_files.append(('examples/ic', glob('./examples/ic/*.py')))
data_files.append(('examples/ic', glob('./examples/ic/Readme')))
'''


package_files = glob('./examples/snapshots/*')
package_files += glob('./config/*parameters')
package_files += glob('./config/rgb_tables/*')
package_files += glob('./config/formats/*.py')
package_files += glob('./config/extensions/*.py')
package_files += glob('./config/opt/*')
package_files += glob('./config/opt/SSP/*')
package_files += glob('./plugins/*.py')
package_files += glob('./fonts/*')
package_files += glob('./config/opt/filters/*')


# DO NOT TOUCH BELOW

# add options
for k in options.keys():
    if options[k]:
        modules_name.append(k)


# Generate extensions
ext_modules = []

for k in modules_name:
    # get flags
    extra_flags = deepcopy(flags)
    if k in spec_flags:
        extra_flags.extend(spec_flags[k])
    # get libraries
    extra_lib = deepcopy(libraries)
    if k in spec_lib:
        extra_lib.extend(spec_lib[k])
    # get library directories
    extra_lib_dirs = []
    if k in spec_lib_dir:
        extra_lib_dirs.extend(spec_lib_dir[k])

    # compile extension
    tmp = Extension("pNbody." + k,
                    glob("src/" + k + "/*.c"),
                    include_dirs=include_dirs,
                    libraries=extra_lib,
                    extra_compile_args=extra_flags,
                    library_dirs=extra_lib_dirs)

    ext_modules.append(tmp)


# add new C extension, but a cleaner way
ext = Extension("pNbody.rtlib",
                glob("src/rtlib/*.c"),
                include_dirs=include_dirs+["src/rtlib/include"],
                libraries=extra_lib,
                extra_compile_args=extra_flags,
                library_dirs=extra_lib_dirs)

ext_modules.append(ext)


# scripts to install
scripts = glob("scripts/pNbody/*")
scripts += glob("scripts/mockimgs/*")
scripts += glob("scripts/imf/*")
scripts += glob("scripts/isochrones/*")
scripts += glob("scripts/ic/*")
scripts += glob("scripts/plot/*")
scripts += glob("scripts/orbits/*")
scripts += glob("scripts/rt/*")
scripts += glob("scripts/il/*")
scripts += glob("scripts/sw/*")
scripts += glob("scripts/ssp/*")
scripts += glob("scripts/sed/*")
scripts += glob("scripts/pnbmov/*")
scripts += glob("scripts/snapshots/*")

setup(
    name="pNbody",
    author="Yves Revaz",
    author_email="yves.revaz@epfl.ch",
    url="http://obswww.unige.ch/~revaz/pNbody/index.html",
    description="""
    This module provides lots of tools used
    to deal with Nbody particles models.
    """,
    long_description_content_type="text/x-rst",
    license="GPLv3",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    packages=find_packages(),
    scripts=scripts,
    ext_modules=ext_modules,
    # data_files=data_files,
    package_data={'pNbody': package_files},
    install_requires=["wheel","setuptools","scipy", "h5py","astropy", "pillow", "tqdm", "versioneer", "ipython"],

)



#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      phot.py
#  brief:     Photometry functions
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################


import numpy as np
from . import ctes
from . import units


try:
    from scipy import optimize
    is_scipy = True
except ImportError:
    is_scipy = False


##########################################################################
#
# SOME FUNCTIONS RELATED TO PHOTOMETRY / OBSERVATION
#
##########################################################################

def MvtoLv(Mv, Mvsun=4.83, Lvsun=1.0):
    """
    Mv    : magnitude in V-band
    Mvsun : magnitude of the sun in V-band
    Lvsun : V-band solar luminosity in solar luminosity unit

    Return the corresponding V-band luminosity in solar luminosity unit.
    """

    Lv = Lvsun * 10**((Mvsun - Mv) / 2.5)

    return Lv


def LvtoMv(Lv, Mvsun=4.83, Lvsun=1.0):
    """
    Lv    : V-band luminosity in solar luminosity unit
    Mvsun : magnitude of the sun in V-band
    Lvsun : V-band solar luminosity in solar luminosity unit

    Return the magnitude in V-band
    """

    Mv = Mvsun - 2.5 * np.log10(Lv / Lvsun)

    return Mv



#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      libutil.py
#  brief:     utilities function (contains plot functions)
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################


import types

from copy import deepcopy

import numpy as np
from pNbody.nbodymodule import *
from pNbody import mapping
from .palette import *
from pNbody.myNumeric import *
import pNbody

try:
    import tkinter as tk
    from PIL import ImageTk
    is_tk = True
except ImportError:
    is_tk = False


from PIL import Image
from PIL import ImageDraw
from PIL import ImageFont
from PIL import ImagePalette


from . import mpiwrapper as mpi
from . import param
import _thread

try:
    from scipy import ndimage
    from scipy import convolve as conv
    is_scipy = True
except ImportError:
    is_scipy = False


try:
    from . import libqt
    is_qt = True
except BaseException:
    is_qt = False


def get_fake_fct(nx=256, ny=256):
    # create a function
    mx = 2 * np.pi
    my = 2 * np.pi
    x, y = np.indices((nx, ny))
    x = mx * (x - nx / 2) / (nx / 2)
    y = my * (y - ny / 2) / (ny / 2)

    R = np.sqrt(x**2 + y**2)
    R1 = np.sqrt((x - np.pi)**2 + (y - np.pi)**2)
    mat = np.cos(R) / (1 + R) + 0.5 * np.cos(R1) / (1 + R1)
    return mat


def get_n_per_task(ntot, NTask):

    n_per_task = np.zeros(NTask, int)
    for Task in range(NTask - 1, -1, -1):
        n_per_task[Task] = ntot / NTask + ntot % NTask * (Task == 0)

    return n_per_task


def get_npart_all(npart, NTask):

    npart_all = np.zeros((NTask, len(npart)), int)

    i = 0
    for n in npart:
        npart_all[:, i] = get_n_per_task(n, mpi.mpi_NTask())
        i = i + 1

    return npart_all


def old_get_n_per_task(ntot, NTask):

    nleft = ntot
    n_per_task = np.zeros(NTask, int)

    for Task in range(NTask - 1, -1, -1):
        if nleft < 2 * ntot / NTask:
            n_per_task[Task] = nleft
        else:
            n_per_task[Task] = ntot / NTask
        nleft = nleft - n_per_task[Task]

    if nleft != 0:
        raise Exception("get_n_per_task : nleft != 0 (%d)" % (nleft))

    return n_per_task


def cross_product(x, y):

    if x.shape != y.shape:
        raise Exception(
            "cross_product error",
            "x and y do not have the same shape")

    n = len(x)
    p = np.zeros((n, 3), x.dtype)

    p[:, 0] = x[:, 1] * y[:, 2] - x[:, 2] * y[:, 1]
    p[:, 1] = x[:, 2] * y[:, 0] - x[:, 0] * y[:, 2]
    p[:, 2] = x[:, 0] * y[:, 1] - x[:, 1] * y[:, 0]

    return p


def myhistogram(a, bins):
    """
    Return the histogram (n x 1 float array) of the
    n x 1 array "a".
    "bins" (m x 1 array) specify the bins of the histogram.
    """
    n = np.searchsorted(np.sort(a), bins)
    n = np.concatenate([n, [len(a)]])
    return n[1:] - n[:-1]


def compress_from_lst(x, num, lst, reject=False):
    """
    Return the compression of x
    """

    # 1) sort the list
    ys = np.sort(lst)

    # 2) sort index in current file
    xs = np.sort(num)
    # sort 0,1,2,n following xs (or self.num)
    zs = np.take(np.arange(len(x)), np.argsort(num))

    # 3) apply mask on sorted lists (here, getmask need xs and ys to be sorted)
    m = np.getmask(xs.astype(int), ys.astype(int))
    if reject:
        m = np.logical_not(m)

    # 4) revert mask, following zs inverse transformation
    c = np.take(m, np.argsort(zs))

    return np.compress(c, x, axis=0)


def tranfert_functions(rmin, rmax, g=None, gm=None):
    """
    This function computes the normalized transfer function from g and gm
    It is very useful to transform a linear vector in a non linear one

    example of g:

      g  = lambda r:np.log(r/rc+1)
      gm = lambda r:rc*(np.exp(r)-1)

    """

    rmin = float(rmin)
    rmax = float(rmax)

    def f(r): return r	 # by default, identity

    def fm(r): return r	 # by default, identity

    if g is not None and gm is not None:
        def f(r): return (g(r) - g(rmin)) / \
            (g(rmax) - g(rmin)) * (rmax - rmin) + rmin

        def fm(f): return gm((f - rmin) *
                             (g(rmax) - g(rmin)) / (rmax - rmin) + g(rmin))

    return f, fm


def geter(n, rmin, rmax, g, gm):
    """
    Generate a one dimentional non linear array of r
    """

    n = int(n)
    dr = (rmax - rmin) / float(n - 1)

    f, fm = tranfert_functions(rmin, rmax, g, gm)

    Rs = np.arange(0, rmax + dr, dr)
    Rs = fm(Rs)

    return Rs


def geter2(n, rmin, rmax, g, gm):
    """
    Generate a one dimentional non linear array of r
    """

    n = int(n)
    dr = (rmax - rmin) / float(n - 1)

    f, fm = tranfert_functions(rmin, rmax, g, gm)

    Rs = np.arange(0, rmax + dr, dr)
    Rs = f(Rs)

    return Rs


def getr(nr=31, nt=32, rm=100.):
    """
    Return a sequence of number (n x 1 array),
    where n=nr+1 defined by: Pfenniger & Friedli (1994)
    """
    j = np.arange(nr + 1)
    ct1 = 0.5 + (nt / (np.pi + np.pi))
    ct2 = (np.exp((nr) / ct1) - 1.)
    r = rm * (np.exp(j / ct1) - 1) / ct2
    return r


def invgetr(r, nr=31, nt=32, rm=100.):
    """
    From r, return the corresponding indexes.
    Inverse of getr function.
    """

    ct1 = 0.5 + (nt / (np.pi + np.pi))
    ct2 = (np.exp((nr) / ct1) - 1.)

    i = ct1 * np.log(ct2 * r / rm + 1)

    return i


def get_eyes(x0, xp, alpha, dr):
    """
    Return the position of two eyes.

    x0 	: position of the head
    xp  	: looking point
    theta : rotation of the head
    dr    : distance of the eyes

    """

    x0 = np.array(x0, float)
    xp = np.array(xp, float)

    # distance between the observer and the looking point
    Robs = np.sqrt((x0[0] - xp[0])**2 + (x0[1] - xp[1])**2 + (x0[2] - xp[2])**2)

    # init
    x = np.array([[dr, -Robs, 0.], [-dr, -Robs, 0.]], np.float32)
    m = pNbody.Nbody(pos=x, status='new')

    # first : put xp in the origin
    x0 = x0 - xp
    # angle in azimuth
    phi = np.arctan2(x0[1], x0[0]) + np.pi / 2
    # angle in nutation
    R = np.sqrt(x0[0]**2 + x0[1]**2)
    if R == 0:
        if x0[2] >= 0:
            theta = np.pi
        else:
            theta = -np.pi
    else:
        theta = np.arctan(x0[2] / R)

    # rotate
    # rotations alpha
    if alpha != 0.:
        m.rotate(alpha, axis='y', mode='p')
    # rotation in nutation
    if theta != 0.:
        m.rotate(-theta, axis='x', mode='p')
    # rotation in azimuth
    if phi != 0.:
        m.rotate(phi, axis='z', mode='p')
    # translate
    m.translate(xp)

    x1 = m.pos[0, :]
    x2 = m.pos[1, :]

    return x1, x2


def apply_filter(mat, name=None, opt=None):
    """
    Apply a filter to an image
    """

    if name == "convol":

        nx = opt[0]
        ny = opt[1]
        sx = float(opt[2])
        sy = float(opt[3])

        nxd = int((nx - 1) / 2)
        nyd = int((ny - 1) / 2)

        # the kernel
        b = np.fromfunction(lambda j, i: np.exp(-(i - nxd)**2 / sx**2 + -(j - nyd)**2 / sy**2), (nx, ny))
        s = np.sum(sum(b))
        b = b / s				  # normalisation

        # conversion:
        b = b.astype(float)
        mat = mat.astype(float)

        return np.convol(mat, b)

    elif name == "convolve":
        nx = opt[0]
        ny = opt[1]
        sx = float(opt[2])
        sy = float(opt[3])

        nxd = int((nx - 1) / 2)
        nyd = int((ny - 1) / 2)

        # the kernel
        b = np.fromfunction(lambda j, i: np.exp(-(i - nxd)**2 / sx**2 + -(j - nyd)**2 / sy**2), (nx, ny))
        s = np.sum(sum(b))
        b = b / s				 # normalisation

        # conversion:
        b = b.astype(float)
        mat = mat.astype(float)

        return conv.convolve2d(mat, b, output=None, fft=0)

    if name == "boxcar":
        nx = opt[0]
        ny = opt[1]
        boxshape = (nx, ny)
        return conv.boxcar(mat, boxshape, mode='reflect')

    if name == "gaussian":
        sigma = float(opt[0])
        return ndimage.gaussian_filter(mat, sigma, mode='nearest')

    if name == "uniform":
        sigma = float(opt[0])
        return ndimage.uniform_filter(mat, sigma, mode='nearest')

    elif name is None:
        pass
    else:
        print("unknown filter type")
        return mat


def set_ranges(mat, scale='log', mn=None, mx=None, cd=None):
    """
    Transform an n x m float array into an n x m int array that will be
    used to create an image. The float values are rescaled and cutted in order to range
    between 0 and 255.

    mat   : the matrice
    scale	: lin or log
    mn  	: lower value for the cutoff
    mx 	: higer value for the cutoff
    cd	: parameter

    """

    rm = np.ravel(mat)

    if mn is None:
        mn = min(rm)
    if mx is None:
        mx = max(rm)

    if mn == mx:
        mn = min(rm)
        mx = max(rm)

    mat = np.clip(mat, mn, mx)

    if scale == 'log':

        if cd is None or cd == 0:
            cd = rm.mean()

        try:
            mat = 255. * np.log(1. + (mat - mn) / (cd)) / \
                np.log(1. + (mx - mn) / (cd))
        except BaseException:
            mat = mat * 0.

    elif scale == 'lin':

        mat = 255. * (mat - mn) / (mx - mn)
        cd = 0

    return mat.astype(int), mn, mx, cd


def contours(m, matint, nl, mn, mx, kx, ky, color, crush='no'):
    """

    Compute iso-contours on a n x m float array.
    If "l_min" equal "l_max", levels are automatically between the minimum and
    maximum values of the matrix "mat".


    m 		= matrice (real values)
    matint	= matrice (interger values)
    kx		= num of sub-boxes
    ky		= num of sub-boxes
    nl		= # of levels
    mn		= min
    mx 		= max
    color		= color of contours

    """

    # create an empty matrix
    newm = np.zeros(m.shape, np.float32)

    if color != 0:

        # transform color
        color = np.array(color, np.int8)

        if mx == mn:
            rm = np.ravel(m)
            mn = min(rm)
            mx = max(rm)
            levels = np.arange(mn + (mx - mn) / nl, mx, (mx - mn) / (nl + 1))
        else:
            levels = np.arange(mn,
                            mx + (mx - mn) / (nl - 1),
                            (mx - mn) / (nl - 1))[:nl]

        # print levels

        m = np.transpose(m)		  # !!!

        X = [(), (), (), (), ()]
        rect = [(1, 2), (2, 3), (3, 4), (4, 1)]

        # number of sub-boxes per axis
        nx = int((m.shape[0] - 1) / (kx - 1))
        ny = int((m.shape[1] - 1) / (ky - 1))
      

        for i in range(0, nx):
            for j in range(0, ny):

                ix = (kx - 1) * i		  # here, we could add an offset
                iy = (ky - 1) * j

                X[1] = (ix, iy, m[iy][ix])
                X[2] = (ix + (kx - 1), iy, m[iy][ix + (kx - 1)])
                X[3] = (ix + (kx - 1), iy + (ky - 1),
                        m[iy + (ky - 1)][ix + (kx - 1)])
                X[4] = (ix, iy + (ky - 1), m[iy + (ky - 1)][ix])

                # find the center
                X[0] = (ix + 0.5 * kx, iy + 0.5 * ky, 0.25 *
                        (X[1][2] + X[2][2] + X[3][2] + X[4][2]))

                # loop over the levels
                for l in levels:
                    # loop over the triangles

                    for r in rect:

                        z1 = X[r[0]][2]
                        z2 = X[r[1]][2]
                        z3 = X[0][2]

                        # find the maximum

                        if z1 > z2:
                            if z1 > z3:
                                if z2 > z3:
                                    c = X[r[0]]
                                    b = X[r[1]]
                                    a = X[0]
                                else:
                                    c = X[r[0]]
                                    b = X[0]
                                    a = X[r[1]]

                            else:
                                c = X[0]
                                b = X[r[0]]
                                a = X[r[1]]

                        else:
                            if z2 > z3:
                                if z1 > z3:
                                    c = X[r[1]]
                                    b = X[r[0]]
                                    a = X[0]
                                else:
                                    c = X[r[1]]
                                    b = X[0]
                                    a = X[r[0]]
                            else:
                                c = X[0]
                                b = X[r[1]]
                                a = X[r[0]]

                        # a,b,c are the tree triangle points

                        # create_line(newm,a[0],a[1],b[0],b[1],color)
                        # create_line(newm,b[0],b[1],c[0],c[1],color)
                        # create_line(newm,c[0],c[1],a[0],a[1],color)

                        # the level cut the triangle
                        if l >= a[2] and l <= c[2] and a[2] != c[2]:

                            lamb = (l - a[2]) / (c[2] - a[2])

                            xx1 = int(lamb * (c[0] - a[0]) + a[0])
                            yy1 = int(lamb * (c[1] - a[1]) + a[1])

                            if l >= b[2] and l <= c[2] and b[2] != c[2]:

                                lamb = (l - b[2]) / (c[2] - b[2])
                                xx2 = int(lamb * (c[0] - b[0]) + b[0])
                                yy2 = int(lamb * (c[1] - b[1]) + b[1])

                            elif b[2] != a[2]:

                                lamb = (l - a[2]) / (b[2] - a[2])
                                xx2 = int(lamb * (b[0] - a[0]) + a[0])
                                yy2 = int(lamb * (b[1] - a[1]) + a[1])

                                mapping.create_line(newm, xx1, yy1, xx2, yy2, color)

    matc = newm.astype(np.int8)

    if crush == 'yes':
        matint = np.ones(matc.shape)
        matint = matint * 255
        matint = np.where(matc, matc, matint)
    else:
        matint = np.where(matc, matc, matint)

    return matint


def add_box(
    matint, shape=(
        256, 256), size=(
            30., 30.), center=None, box_opts=(
                1, None, None, 255)):
    """
    add a box on the frame
    """

    lweight = box_opts[0]
    xticks = box_opts[1]
    yticks = box_opts[2]
    color = box_opts[3]

    box = sbox(
        shape,
        size,
        lweight=lweight,
        xticks=xticks,
        yticks=yticks,
        color=color)
    matint = np.where(box != 0, box, matint)

    return matint


def mplot(mat, palette='light', save=None, marker=None, header=None):
    """
    plot a 2d array
    """

    from pNbody import iofunc as pnio

    if mpi.mpi_IsMaster():

        if save is not None:
            if os.path.splitext(save)[1] == ".fits":
                pnio.WriteFits(np.transpose(mat).astype(np.float32), save, header)
                return

        # if the matrix is real
        if mat.dtype != np.int32:
            matint, mn_opt, mx_opt, cd_opt = set_ranges(
                mat, scale='lin', cd=0, mn=0, mx=0)
            mat = matint

        # add marker
        if marker is not None:
            if marker == 'cross':
                nx, ny = mat.shape
                ix, iy = np.indices(mat.shape)
                mat = np.where((ix == nx / 2) + (ix == nx / 2 - 1) +
                            (iy == ny / 2) + (iy == ny / 2 - 1), 255, mat)

            if marker == 'circle':
                nx, ny = mat.shape
                ix, iy = np.indices(mat.shape)
                r = 100
                dr = 1
                Rs = np.sqrt((ix - nx / 2)**2 + (iy - ny / 2)**2)
                c = (Rs < r + dr) * (Rs > r - dr)

                mat = np.where(c, 255, mat)

        image = get_image(mat, name=None, palette_name=palette)

        if save is None:
            display(image)
        else:
            image.save(save)


def get_image(mat, name=None, palette_name='light', mode='RGB'):
    """
    Return an image (PIL object).

    data : numpy 2x2 object
    name : name of the output
    palette_name : name of a palette
    """

    # modif 09.03.05
    mat = np.transpose(mat)
    mat = mat.astype(np.int8)

    # mat = mat.astype(int)      # ok
    # mat = np.transpose(mat)        # la transposee fait changer aussi autre
    # chose ??? c'est peut-etre la sortie de Py_BuildValue("(Of)",mat,cdopt);
    # pas bon...

    image = Image.frombytes("P", (mat.shape[1], mat.shape[0]), mat.tostring())

    # include the palette
    palette = Palette(palette_name)
    image.putpalette(palette.palette)

    # set mode
    if mode == 'RGB':			# to convert with ImageQt, need to be in RGB
        image = image.convert('RGB')

    # save it
    if name is not None:
        image.save(name)

    return image


def display(image):

    if mpi.mpi_IsMaster():

        if is_qt:
            libqt.display(image)

        elif is_tk:

            #root = tk.Tk()
            root = tk.Toplevel()
            canvas = tk.Canvas(root)
            canvas.pack()
            imagetk = ImageTk.PhotoImage(image)
            canvas.config(width=image.size[0], height=image.size[1])
            canvas.create_image(0., 0., anchor=tk.NW, image=imagetk)
            root.mainloop()

        else:

            raise Exception("tk or qt not present")


def sbox(shape, size, lweight=1, xticks=None, yticks=None, color=255):
    """
    simple box

    return a matrix of integer, containing a box with labels

    xticks = (m0,d0,h0,m1,d1,h1)

    0 = big
    1 = small

    m0,m1 = dist between ticks
    d0,d1 = first tick
    h0,h1 = height of the ticks

    """
    center = None

    # parameters
    nn = 4.
    bticks = np.array([1., 2., 5.])

    color = int(color)

    # create matrix from scratch
    matint = np.zeros(shape, int)

    # add the outer box
    for l in range(lweight):
        # in x
        matint[:, l] = (np.zeros((shape[0],)) + color).astype(int)
        # may be commented
        matint[:, shape[1] - 1 - l] = (np.zeros((shape[0],)) + color).astype(int)
        # in y
        matint[l, :] = (np.zeros((shape[1],)) + color).astype(int)
        # may be commented
        matint[shape[0] - 1 - l, :] = (np.zeros((shape[1],)) + color).astype(int)

    # add the ticks in x
    if xticks is None:
        #cx = center[0]
        #cy = center[1]
        cx = 0
        cy = 0
        hx = shape[0]
        hy = shape[1]
        dx = size[0]
        dy = size[1]

        rat = (2. * dx) / nn					# dist between ticks
        ratn = rat / 10**int(np.log10(rat))
        d0 = bticks[np.argmin(np.fabs(bticks - ratn))] * 10**int(np.log10(rat))

        n0 = int(np.fmod(2. * dx / d0, 2) + (2. * dx / d0))		# number of ticks
        m0 = cx - (n0 * d0) / 2. 				# first tick
        h0 = hy / 20.						# height of the ticks

        d1 = d0 / 4.						# dist between ticks
        n1 = n0 * 4
        m1 = m0						# first tick
        h1 = h0 / 2.						# height of the ticks

    else:
        m0 = xticks[0]
        d0 = xticks[1]
        h0 = xticks[2]
        m1 = xticks[3]
        d1 = xticks[4]
        h1 = xticks[5]

    # big ticks x
    matint = drawxticks(matint, m0, d0, n0, h0, shape, size, center, color)
    # small ticks x
    matint = drawxticks(matint, m1, d1, n1, h1, shape, size, center, color)

    # print "dx = %8.3f"%(d0)

    # add the ticks in y
    if xticks is None:
        #cx = center[0]
        #cy = center[1]
        cx = 0
        cy = 0
        hx = shape[0]
        hy = shape[1]
        dx = size[0]
        dy = size[1]

        rat = (2. * dx) / nn					# dist between ticks
        ratn = rat / 10**int(np.log10(rat))
        d0 = bticks[np.argmin(np.fabs(bticks - ratn))] * 10**int(np.log10(rat))

        n0 = int(np.fmod(2. * dx / d0, 2) + (2. * dx / d0))		# number of ticks
        m0 = cy - (n0 * d0) / 2. 				# first tick
        h0 = hx / 20.						# height of the ticks

        d1 = d0 / 4.						# dist between ticks
        n1 = n0 * 4
        m1 = m0						# first tick
        h1 = h0 / 2.						# height of the ticks

    else:
        m0 = xticks[0]
        d0 = xticks[1]
        h0 = xticks[2]
        m1 = xticks[3]
        d1 = xticks[4]
        h1 = xticks[5]

    # big ticks y
    matint = drawyticks(matint, m0, d0, n0, h0, shape, size, center, color)
    # small ticks y
    matint = drawyticks(matint, m1, d1, n1, h1, shape, size, center, color)

    # print "dy = %8.3f"%(d0)

    return matint


def drawxticks(matint, m0, d0, n0, h0, shape, size, center, color):
    """
    draw x ticks in a matrix
    """

    x = m0
    for nx in range(n0):
        ix, iy = phys2img(shape, size, center, x, 0)
        try:  # !! version dependant !!
            ix = int(ix[0])
        except BaseException:
            ix = int(ix)

        for iy in range(int(h0)):
            matint[ix, iy] = np.array([color], int)[0]
            matint[ix, shape[0] - 1 - iy] = np.array([color], int)[0]

        x = x + d0

    return matint


def drawyticks(matint, m0, d0, n0, h0, shape, size, center, color):
    """
    draw x ticks in a matrix
    """

    x = m0
    for nx in range(n0):
        ix, iy = phys2img(shape, size, center, x, 0)
        try:  # !! version dependant !!
            ix = int(ix[0])
        except BaseException:
            ix = int(ix)

        for iy in range(int(h0)):
            matint[iy, ix] = np.array([color], int)[0]
            matint[shape[1] - 1 - iy, ix] = np.array([color], int)[0]

        x = x + d0

    return matint


def draw_line(m, x0, x1, y0, y1, c, z0=None, z1=None):

    dd = 0.5  # half a pixel

    imax = m.shape[0] - 1
    jmax = m.shape[1] - 1

    if abs(x0 - x1) > abs(y0 - y1):
        ixmin = int(round(min(x0, x1)))
        ixmax = int(round(max(x0, x1)))

        a = (y1 - y0) / (x1 - x0)
        #az = (z1-z0)/(x1-x0)

        for i in range(ixmin, ixmax + 1):
            j = round((i - x0) * a + y0)
            #jz =       (i - x0)*az + z0

            if (i <= imax) and (i >= 0) and (j <= jmax) and (j >= 0):
                m[i, j] = c

    elif abs(x0 - x1) <= abs(y0 - y1) and abs(y0 - y1) > 0:
        iymin = int(round(min(y0, y1)))
        iymax = int(round(max(y0, y1)))

        a = (x1 - x0) / (y1 - y0)
        #az = (z1-z0)/(y1-y0)

        for j in range(iymin, iymax + 1):
            i = round((j - y0) * a + x0)
            #iz =       (j - y0)*az + z0

            if (i <= imax) and (i >= 0) and (j <= jmax) and (j >= 0):
                m[i, j] = c

    else:
        i = round(x0)
        j = round(y0)
        if (i <= imax) and (i >= 0) and (j <= jmax) and (j >= 0):
            m[i, j] = c

    return m


def draw_points(m, x, y, c):
    n = len(x)
    for i in range(n):
        m = draw_line(m, x[i], x[i], y[i], y[i], c)
    return m


def draw_lines(m, x, y, c):
    n = len(x)
    for i in range(n - 1):
        m = draw_line(m, x[i], x[i + 1], y[i], y[i + 1], c)
    return m


def draw_polygon(m, x, y, c):
    n = len(x)
    for i in range(n - 1):
        m = draw_line(m, x[i], x[i + 1], y[i], y[i + 1], c)
    m = draw_line(m, x[n - 1], x[0], y[n - 1], y[0], c)
    return m


def draw_segments(m, x, y, c, zp=None):
    n = len(x)
    p = int(n / 2)

    for j in range(p):
        i = 2 * j
        if zp is not None:
            m = draw_line(m, x[i], x[i + 1], y[i],
                          y[i + 1], c, zp[i], zp[i + 1])
        else:
            m = draw_line(m, x[i], x[i + 1], y[i], y[i + 1], c)
    return m


def draw_polygonN(m, x, y, c, N):

    n = len(x)
    p = int(n / N)

    for j in range(p):
        i = N * j
        for k in range(N - 1):
            m = draw_line(m, x[i + k], x[i + k + 1], y[i + k], y[i + k + 1], c)
        m = draw_line(m, x[i + N - 1], x[i], y[i + N - 1], y[i], c)

    return m


def phys2img(shape, size, center, x, y):
    """
    convert physical position into the image pixel
    """
    #cx = center[0]
    #cy = center[1]
    cx = 0
    cy = 0
    hx = shape[0]
    hy = shape[1]
    dx = size[0]
    dy = size[1]

    ax = hx / (2. * dx)
    bx = (hx / 2.) * (1. - cx / dx)

    ay = hy / (2. * dy)
    by = (hy / 2.) * (1. - cy / dy)

    ix = int(ax * x + bx)
    iy = int(ay * y + by)

    ix = np.clip(ix, 0, hx - 1)
    iy = np.clip(iy, 0, hy - 1)

    return ix, iy


#################################
def can_to_carth(l, scale_fact, x_can, y_can):
    #################################

    x = float(x_can - l[0] / 2.) / scale_fact[0]
    y = float(-y_can + l[1] / 2.) / scale_fact[1]

    return x, y

#################################


def carth_to_can(l, scale_fact, x, y):
    #################################

    x_can = int(x * scale_fact[0] + l[0] / 2.)
    y_can = int(-y * scale_fact[1] + l[1] / 2.)

    return x_can, y_can

#################################


def getval(nb, mode='m', obs=None):
    #################################
    """
    For each point, return a specific value linked to this point

    Parameters 
    ----------

    mode : str
        m           : moment 0

        0           : moment 0
    
        x           : first moment in x

        y           : first moment in y

        z           : first moment in z
        

        y2  	: second moment in x

        y2  	: second moment in y

        z2  	: second moment in z

        vx		: first velocity moment in x

        vy		: first velocity moment in y

        vz		: first velocity moment in z

        vx2		: second velocity moment in x

        vy2		: second velocity moment in y
        
        vz2		: second velocity moment in z


        Lx   	: kinetic momemtum in x

        Ly   	: kinetic momemtum in y

        Lz   	: kinetic momemtum in z

        lx   	: specific kinetic momemtum in x

        ly   	: specific kinetic momemtum in y
        
        lz   	: specific kinetic momemtum in z


        u		: specific energy

        rho		: density

        T   	: temperature

        A   	: entropy
        
        P   	: pressure

        Tcool       : cooling time

        Lum         : luminosity

        Ne          : local electro density



        r		: first momemtum of radial distance [depends on projection]
        
        r2		: second momemtum of radial distance [depends on projection]

        vr		: first momemtum of radial velocity [depends on projection]

        vr2		: second momemtum of radial velocity [depends on projection]

        vxyr	: first momemtum of radial velocity in the plane [depends on projection]

        vxyr2	: second momemtum of radial velocity in the plane [depends on projection]

        vtr  	: first momemtum of tangential velocity in the plane [depends on projection]

        vtr2	: second momemtum of tangential velocity in the plane [depends on projection]



    Notes
    ------------------

    Three types of values:

    (1) scalar values

        pos,m,tem,u,rho,ne
        vx,vx2,vy,vy2,vz,vz2,vxyr,vxyr2,vxyt,vxyt2

    (2) values computed with respect to xp

        z,z2
        Lx,Ly,Lz,lx,ly,ly


    (3) values computed with respect to absolute position in space

        empty for the moment (do not have the initial positions...)


    """

    ######################################
    # values independent of angle of view
    ######################################

    ###########
    # scalar
    ###########

    # moment 0
    if mode == '0':
        val = np.ones(len(nb.pos))

    # moment 0
    elif mode == 'm':
        val = np.ones(len(nb.pos))

    # u
    elif mode == 'u':
        val = nb.U()

    # rho
    elif mode == 'rho':
        val = nb.Rho()

    # T
    elif mode == 'T':
        val = nb.T()

    # A
    elif mode == 'A':
        val = nb.A()

    # P
    elif mode == 'P':
        val = nb.P()

    # Tcool
    elif mode == 'Tcool':
        val = nb.Tcool()

    # Ne
    elif mode == 'Ne':
        val = nb.Ne()

    # Hsml
    elif mode == 'Hsml':
        val = nb.Hsml()

    # Lum
    elif mode == 'Lum':
        val = nb.Lum()

    ###########
    # vectors
    ###########

    # x
    elif mode == 'x':
        val = nb.pos[:, 0]

    # y
    elif mode == 'y':
        val = nb.pos[:, 1]

    # z
    elif mode == 'z':
        val = nb.pos[:, 2]

    # x2
    elif mode == 'x2':
        val = nb.pos[:, 0]**2

    # y2
    elif mode == 'y2':
        val = nb.pos[:, 1]**2

    # z2
    elif mode == 'z2':
        val = nb.pos[:, 2]**2

    # vx
    elif mode == 'vx':
        val = nb.vel[:, 0]

    # vy
    elif mode == 'vy':
        val = nb.vel[:, 1]

    # vz
    elif mode == 'vz':
        val = nb.vel[:, 2]

    # vx2
    elif mode == 'vx2':
        val = nb.vel[:, 0]**2

    # vy2
    elif mode == 'vy2':
        val = nb.vel[:, 1]**2

    # vz2
    elif mode == 'vz2':
        val = nb.vel[:, 2]**2

    # kinetic momentum in x
    elif mode == 'Lx':
        val = nb.mass * (nb.pos[:, 1] * nb.vel[:, 2] -
                         nb.pos[:, 2] * nb.vel[:, 1])

    # kinetic momentum in y
    elif mode == 'Ly':
        val = nb.mass * (nb.pos[:, 2] * nb.vel[:, 0] -
                         nb.pos[:, 0] * nb.vel[:, 2])

    # kinetic momentum in z
    elif mode == 'Lz':
        val = nb.mass * (nb.pos[:, 0] * nb.vel[:, 1] -
                         nb.pos[:, 1] * nb.vel[:, 0])

    # specific kinetic momentum in x
    elif mode == 'lx':
        val = (nb.pos[:, 1] * nb.vel[:, 2] - nb.pos[:, 2] * nb.vel[:, 1])

    # specific kinetic momentum in y
    elif mode == 'ly':
        val = (nb.pos[:, 2] * nb.vel[:, 0] - nb.pos[:, 0] * nb.vel[:, 2])

    # specific kinetic momentum in z
    elif mode == 'lz':
        val = (nb.pos[:, 0] * nb.vel[:, 1] - nb.pos[:, 1] * nb.vel[:, 0])

    # luminosity
    elif mode == 'lum':
        val = nb.luminosity_spec()

    ########################################
    # values dependent on the angle of view
    ########################################

    # first moment in z
    elif mode == 'r':
        # center xp
        nb.translate(obs[0] - obs[1])
        val = nb.pos[:, 2]

    # second moment in z
    elif mode == 'r2':
        # center xp
        nb.translate(obs[0] - obs[1])
        val = nb.pos[:, 2]**2

    # first moment in z
    elif mode == 'vr':
        val = nb.vel[:, 2]

    # second moment in z
    elif mode == 'vr2':
        # center xp
        val = nb.vel[:, 2]**2

    # first velocity moment in radial velocity in the plane
    elif mode == 'vxyr':
        val = (nb.pos[:, 0] * nb.vel[:, 0] + nb.pos[:, 1] *
               nb.vel[:, 1]) / np.sqrt(nb.pos[:, 0]**2 + nb.pos[:, 1]**2)

    # second velocity moment in radial velocity in the plane
    elif mode == 'vxyr2':
        val = ((nb.pos[:, 0] * nb.vel[:, 0] + nb.pos[:, 1] *
                nb.vel[:, 1]) / np.sqrt(nb.pos[:, 0]**2 + nb.pos[:, 1]**2))**2

    # first moment in tangential velocity in the plane
    elif mode == 'vtr':
        val = (nb.pos[:, 0] * nb.vel[:, 1] - nb.pos[:, 1] *
               nb.vel[:, 0]) / np.sqrt(nb.pos[:, 0]**2 + nb.pos[:, 1]**2)

    # second moment in tangential velocity in the plane
    elif mode == 'vtr2':
        val = ((nb.pos[:, 0] * nb.vel[:, 1] - nb.pos[:, 1] *
                nb.vel[:, 0]) / np.sqrt(nb.pos[:, 0]**2 + nb.pos[:, 1]**2))**2

    # other mode
    else:
        # print "getval : unknown mode %s"%(mode)
        val = eval(mode)

    del nb

    return val.astype(np.float32)


#################################
def getvaltype(mode='m'):
    #################################
    """
    list values that depends on projection
    """

    if mode == 'r':
        valtype = 'in projection'

    elif mode == 'r2':
        valtype = 'in projection'

    elif mode == 'vr':
        valtype = 'in projection'

    elif mode == 'vr2':
        valtype = 'in projection'

    elif mode == 'vxyr':
        valtype = 'in projection'

    elif mode == 'vxyr2':
        valtype = 'in projection'

    elif mode == 'vtr':
        valtype = 'in projection'

    elif mode == 'vtr2':
        valtype = 'in projection'

    # other mode
    else:
        valtype = 'normal'

    return valtype


#################################
def extract_parameters(arg, kw, defaultparams):
    #################################
    """
    this function extract parameters given to a function
    it returns a dictionary of parameters with respective value

    defaultparams : dictionary of default parameters


    """

    params = {}

    if len(kw) == 0 and len(arg) >= 1:
        if isinstance(arg[0], dict):
            params = arg[0]

    elif len(kw) >= 1 and len(arg) >= 1:

        if isinstance(arg[0], dict):
            params = arg[0]

            # add other keywords
            for key in list(kw.keys()):
                params[key] = kw[key]

    elif len(kw) >= 1 and len(arg) == 0:
        params = kw

    newparams = deepcopy(defaultparams)
    for key in list(params.keys()):
        if key in defaultparams:
            newparams[key] = params[key]

    return newparams


##########################################################
# functions relative to mapping
##########################################################

#################################
def GetNumberMap(pos, shape):
    #################################
    """
    """

    val = np.ones(pos.shape).astype(np.float32)

    if len(shape) == 1:
        # compute zero momentum
        m0 = mapping.mkmap1d(pos, val, val, shape)

    elif len(shape) == 2:
        # compute zero momentum
        m0 = mapping.mkmap2d(pos, val, val, shape)

    elif len(shape) == 3:
        # compute zero momentum
        m0 = mapping.mkmap3d(pos, val, val, shape)

    else:
        return

    return m0

#################################


def GetMassMap(pos, mass, shape):
    #################################
    """
    """

    val = np.ones(pos.shape).astype(np.float32)

    if len(shape) == 1:
        # compute zero momentum
        m0 = mapping.mkmap1d(pos, mass, val, shape)

    elif len(shape) == 2:
        # compute zero momentum
        m0 = mapping.mkmap2d(pos, mass, val, shape)

    elif len(shape) == 3:
        # compute zero momentum
        m0 = mapping.mkmap3d(pos, mass, val, shape)

    else:
        return

    return m0

#################################


def GetMeanValMap(pos, mass, val, shape):
    #################################
    """
    """

    if len(shape) == 1:
        # compute zero momentum
        m0 = mapping.mkmap1d(pos, mass, np.ones(len(pos), np.float32), shape)
        # compute first momentum
        m1 = mapping.mkmap1d(pos, mass, val, shape)

    elif len(shape) == 2:
        # compute zero momentum
        m0 = mapping.mkmap2d(pos, mass, np.ones(len(pos), np.float32), shape)
        # compute first momentum
        m1 = mapping.mkmap2d(pos, mass, val, shape)

    elif len(shape) == 3:
        # compute zero momentum
        m0 = mapping.mkmap3d(pos, mass, np.ones(len(pos), np.float32), shape)
        # compute first momentum
        m1 = mapping.mkmap3d(pos, mass, val, shape)

    else:
        return

    # combi map
    return GetMeanMap(m0, m1)


#################################
def GetSigmaValMap(pos, mass, val, shape):
    #################################
    """
    """

    if len(shape) == 1:
        # compute zero momentum
        m0 = mapping.mkmap1d(pos, mass, np.ones(len(pos), np.float32), shape)
        # compute first momentum
        m1 = mapping.mkmap1d(pos, mass, val, shape)
        # compute second momentum
        m2 = mapping.mkmap1d(pos, mass, val * val, shape)

    elif len(shape) == 2:
        # compute zero momentum
        m0 = mapping.mkmap2d(pos, mass, np.ones(len(pos), np.float32), shape)
        # compute first momentum
        m1 = mapping.mkmap2d(pos, mass, val, shape)
        # compute second momentum
        m2 = mapping.mkmap2d(pos, mass, val * val, shape)

    elif len(shape) == 3:
        # compute zero momentum
        m0 = mapping.mkmap3d(pos, mass, np.ones(len(pos), np.float32), shape)
        # compute first momentum
        m1 = mapping.mkmap3d(pos, mass, val, shape)
        # compute second momentum
        m2 = mapping.mkmap3d(pos, mass, val * val, shape)

    else:
        return

    # combi map
    return GetSigmaMap(m0, m1, m2)


#################################
def GetMeanMap(m0, m1):
    #################################
    """
    Return a MeanMap using the 0 and 1 momentum

    m0 : zero  momentum
    m1 : first momentum
    """
    m1 = np.where(m0 == 0, 0, m1)
    m0 = np.where(m0 == 0, 1, m0)
    mat = m1 / m0

    return mat

#################################


def GetSigmaMap(m0, m1, m2):
    #################################
    """
    Return a MeanMap using the 0 and 1 and 2 momentum

    m0 : zero   momentum
    m1 : first  momentum
    m2 : second momentum
    """
    m1 = np.where(m0 == 0, 0, m1)
    m2 = np.where(m0 == 0, 0, m2)
    m0 = np.where(m0 == 0, 1, m0)
    mat = m2 / m0 - (m1 / m0)**2
    mat = np.where((mat > 0), np.sqrt(mat), 0)

    return mat


##########################################################
# extract from map functions
##########################################################

def Extract1dMeanFrom2dMap(x, y, mass, val, kx, ky, nmin, momentum=0):
    """

    Extract the mean along one axis, from a 2d mean or sigma matrix

    x    : pos in first dim.  (beetween 0 and 1)
    y    : pos in sec   dim.  (beetween 0 and 1)
    mass : mass of particles
    val  : value to compute

    kx : number of bins in x
    ky : number of bins in y

    nmin : min number of particles needed to compute value

    momentum : 0,1,2 (-1=number)
    """

    shape = (kx, ky)

    # normalize values

    pos = np.transpose(np.array([x, y, y], np.float32))

    mat_num = GetNumberMap(pos, shape)

    if momentum == -1:
        mat_val = GetNumberMap(pos, shape)
    if momentum == 0:
        mat_val = GetMassMap(pos, mass, shape)
    elif momentum == 1:
        mat_val = GetMeanValMap(pos, mass, val, shape)
    elif momentum == 2:
        mat_val = GetSigmaValMap(pos, mass, val, shape)

    ################
    # 2d mean
    ################

    mat_num = np.transpose(mat_num)
    mat_val = np.transpose(mat_val)

    m1 = np.sum(mat_num, 0)
    m0 = np.sum(np.ones(mat_val.shape), 0)
    vec_num = np.where((m0 != 0), m1 / m0, 0)

    c = (mat_num > nmin)  # *(mat_val!=0)
    m1 = np.sum(mat_val * c, 0)
    m0 = np.sum(np.ones(mat_val.shape) * c, 0)
    vec_sigma = np.where((m0 != 0), m1 / m0, 0)

    return vec_sigma


def get1dMeanFrom2dMap(mat_val, mat_num, nmin=32, axis=0):

    m1 = np.sum(mat_num, axis)
    m0 = np.sum(np.ones(mat_val.shape), axis)
    vec_num = np.where((m0 != 0), m1 / m0, axis)

    c = (mat_num > nmin)
    m1 = np.sum(mat_val * c, axis)
    m0 = np.sum(np.ones(mat_val.shape) * c, axis)
    vec_sigma = np.where((m0 != 0), m1 / m0, axis)

    return vec_sigma


##########################################################
# filter
##########################################################

#################################
def log_filter(x, xmin, xmax, xc, kx=1.0):
    #################################
    """
    map a value between 0 and kx
    """

    if xc == 0:
        return kx * (x - xmin) / (xmax - xmin)
    else:
        return kx * np.log(1 + (x - xmin) / xc) / np.log(1 + (xmax - xmin) / xc)

#################################


def log_filter_inv(k, xmin, xmax, xc, kx=1.0):
    #################################
    """
    map a value betwen xmin and xmax
    """
    if xc == 0:
        return (k / kx * (xmax - xmin)) + xmin
    else:
        A = np.log(1 + (xmax - xmin) / xc)
        return xc * (np.exp(A * k / kx) - 1.0) + xmin


##########################################################
# change of coordinate
##########################################################


######################
# cylindrical coord
######################

def vel_cyl2cart(pos=None, vel=None):
    """
    Transform velocities in cylindrical coordinates vr,vt,vz into carthesian
    coodinates vx,vy,vz.
    Pos is the position of particles in cart. coord.
    Vel is the velocity in cylindrical coord.
    Return a 3xn float array.
    """

    x = pos[:, 0]
    y = pos[:, 1]
    z = pos[:, 2]

    vr = vel[:, 0]
    vt = vel[:, 1]
    vz = vel[:, 2]

    r = np.sqrt(x**2 + y**2)

    vx = np.where(r > 0, (vr * x - vt * y) / r, 0)
    vy = np.where(r > 0, (vr * y + vt * x) / r, 0)
    vz = vz

    return np.transpose(np.array([vx, vy, vz])).astype(np.float32)


def vel_cart2cyl(pos, vel):
    """
    Transform velocities in carthesian coordinates vx,vy,vz into cylindrical
    coodinates vr,vz,vz.
    Pos is the position of particles in cart. coord.
    Vel is the velocity in cart. coord.
    Return a 3xn float array.
    """

    x = pos[:, 0]
    y = pos[:, 1]
    z = pos[:, 2]

    vx = vel[:, 0]
    vy = vel[:, 1]
    vz = vel[:, 2]

    r = np.sqrt(x**2 + y**2 + z**2)

    vr = np.where(r > 0, (x * vx + y * vy) / r, 0)
    vt = np.where(r > 0, (x * vy - y * vx) / r, 0)
    vz = vz

    return np.transpose(np.array([vr, vt, vz])).astype(np.float32)


##########################################################
# gl2nbody
##########################################################


def RotateAround(angle, axis, point, ObsM):
    """
    this should be C
    """


    # this work with OpenGL
    #Q = np.ones(16,float)
    # glLoadIdentity();
    # glTranslated(point[0],point[1],point[2]);
    # glRotated(angle,axis[0],axis[1],axis[2]);
    # glTranslated(-point[0],-point[1],-point[2]);
    ##Q = glGetDoublev(GL_MODELVIEW_MATRIX);
    # glMultMatrixd(ObsM);
    #ObsM = glGetDoublev(GL_MODELVIEW_MATRIX);
    # return np.ravel(ObsM)

    angle = angle * np.pi / 180
    point = np.concatenate((point, np.array([0])))

    M = ObsM
    M.shape = (4, 4)

    # center point
    M = M - point

    # construction of the rotation matrix
    norm = np.sqrt(axis[0]**2 + axis[1]**2 + axis[2]**2)
    if norm == 0:
        raise ValueError("getting norm=sqrt(ax[0]^2+ax[1]^2+ax[2]^2) = 0, which I can't handle.")
    sn = np.sin(-angle / 2.)

    e0 = np.cos(-angle / 2.)
    e1 = axis[0] * sn / norm
    e2 = axis[1] * sn / norm
    e3 = axis[2] * sn / norm

    a = np.zeros((4, 4), float)
    a[0, 0] = e0**2 + e1**2 - e2**2 - e3**2
    a[1, 0] = 2. * (e1 * e2 + e0 * e3)
    a[2, 0] = 2. * (e1 * e3 - e0 * e2)
    a[3, 0] = 0.
    a[0, 1] = 2. * (e1 * e2 - e0 * e3)
    a[1, 1] = e0**2 - e1**2 + e2**2 - e3**2
    a[2, 1] = 2. * (e2 * e3 + e0 * e1)
    a[3, 1] = 0.
    a[0, 2] = 2. * (e1 * e3 + e0 * e2)
    a[1, 2] = 2. * (e2 * e3 - e0 * e1)
    a[2, 2] = e0**2 - e1**2 - e2**2 + e3**2
    a[3, 2] = 0.
    a[0, 3] = 0.
    a[1, 3] = 0.
    a[2, 3] = 0.
    a[3, 3] = 1.

    a = a.astype(float)

    # multiply x and v
    M = np.dot(M, a)

    # decenter point
    M = M + point

    return np.ravel(M)


def gl2pNbody(glparam, nbparam=None):

    EYE = 0
    PTS = 4
    HEA = 8
    ARM = 12

    gwinShapeX = 512				# to change...
    gwinShapeY = 512

    ProjectionMode = glparam["ProjectionMode"]

    if (ProjectionMode):
        # frustum
        gwinPerspectiveTop = glparam["PerspectiveNear"] * \
            np.tan(glparam["PerspectiveFov"] * 0.5 * np.pi / 180.)
        gwinPerspectiveRight = gwinPerspectiveTop * \
            float(gwinShapeX) / float(gwinShapeY)
    else:
        # ortho
        gwinPerspectiveLeft = -5 * glparam["PerspectiveNear"]
        gwinPerspectiveRight = -gwinPerspectiveLeft
        gwinPerspectiveTop = float(gwinShapeY) / \
            float(gwinShapeX) * gwinPerspectiveRight

    gwinClip1 = glparam["PerspectiveNear"]
    gwinClip2 = glparam["PerspectiveFar"]

    # /*********************/
    # /* observer position */
    # /*********************/

    # /* copy the observer matrix */
    # ObsObject.ComputeEyes();

    M = np.zeros(16, np.float32)
    axis = np.zeros(3, np.float32)
    point = np.zeros(3, np.float32)

    M[0] = glparam["M0"]
    M[1] = glparam["M1"]
    M[2] = glparam["M2"]
    M[3] = glparam["M3"]
    M[4] = glparam["M4"]
    M[5] = glparam["M5"]
    M[6] = glparam["M6"]
    M[7] = glparam["M7"]
    M[8] = glparam["M8"]
    M[9] = glparam["M9"]
    M[10] = glparam["M10"]
    M[11] = glparam["M11"]
    M[12] = glparam["M12"]
    M[13] = glparam["M13"]
    M[14] = glparam["M14"]
    M[15] = glparam["M15"]

    # rotate
    axis[0] = M[EYE + 0] - M[PTS + 0]
    axis[1] = M[EYE + 1] - M[PTS + 1]
    axis[2] = M[EYE + 2] - M[PTS + 2]
    point[0] = M[EYE + 0]
    point[1] = M[EYE + 1]
    point[2] = M[EYE + 2]

    M = RotateAround(90, axis, point, M)			# this is also bad !!!

    # this is not correct, dist must come from projec. params */
    dist = np.sqrt((M[0] - M[8])**2 + (M[1] - M[9])**2 + (M[2] - M[10])**2)

    # head
    obs1 = M[0]
    obs2 = M[1]
    obs3 = M[2]

    # lookat point */
    norm = np.sqrt((M[0] - M[4])**2 + (M[1] - M[5])**2 + (M[2] - M[6])**2)

    obs4 = obs1 + (M[4] - obs1) / norm * dist
    obs5 = obs2 + (M[5] - obs2) / norm * dist
    obs6 = obs3 + (M[6] - obs3) / norm * dist

    # arm
    norm = np.sqrt((M[0] - M[8])**2 + (M[1] - M[9])**2 + (M[2] - M[10])**2)

    obs7 = obs1 + (M[8] - obs1) / norm
    obs8 = obs2 + (M[9] - obs2) / norm
    obs9 = obs3 + (M[10] - obs3) / norm

    # head
    norm = np.sqrt((M[0] - M[12])**2 + (M[1] - M[13])**2 + (M[2] - M[14])**2)

    obs10 = obs1 + (M[12] - obs1) / norm
    obs11 = obs2 + (M[13] - obs2) / norm
    obs12 = obs3 + (M[14] - obs3) / norm

    obs = np.array([obs1, obs2, obs3, obs4, obs5, obs6, obs7,
                 obs8, obs9, obs10, obs11, obs12], float)
    obs.shape = (4, 3)

    if nbparam is None:
        nbparam = param.Params(PARAMETERFILE, None)

    nbparam.set('obs', obs)
    nbparam.set('X0', 'None')
    nbparam.set('xp', 'None')
    nbparam.set('alpha', 'None')
    nbparam.set('view', 'xy')
    nbparam.set('r_obs', dist)
    nbparam.set('clip', (gwinClip1, gwinClip2))
    nbparam.set('eye', 'None')
    nbparam.set('dist_eye', 0.1)
    nbparam.set('foc', 0.1)

    if ProjectionMode:
        nbparam.set('persp', 'on')
        nbparam.set('cut', 'yes')
    else:
        nbparam.set('persp', 'off')
        nbparam.set('cut', 'no')

    nbparam.set('shape', (gwinShapeX, gwinShapeY))
    nbparam.set('size', (gwinPerspectiveRight, gwinPerspectiveTop))

    return nbparam


def get_CommandLine():
  """
  return the command line from invocated
  """
  return " ".join(sys.argv)
  
def get_GitTag():
  """
  return the pNbody git tag
  """
  import pNbody
  return pNbody.__version__

def get_UserName():
  """
  return the user name
  """

  if "USER" in os.environ:
    username      = os.environ['USER']
  else:
    username      = 'unknown'
  
  return username
  

def get_Date():
  """
  get the date
  """
  import time
  return time.ctime()


  
  
  
  
  
    
  
    
  



###########################################################################################
#  package:   pNbody
#  file:      newstar.py
#  brief:
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

from Nbody import *
from Numeric import *

from tkinter import *


class Dissip:

    def __init__(self, master):

        self.master = master

        self.varRad = StringVar()

        self.root = Toplevel()
        box = Frame(self.root)
        box.grid(column=0, row=0, sticky=E + N + S + W)

        Label(box, text='tstar < ').grid(column=0, row=3, sticky=E + N + S + W)

        self.t1Ent = Entry(box)
        self.t1Ent.grid(column=1, row=3, sticky=E + N + S + W)

        but = Frame(self.root)
        but.grid(column=0, row=1, sticky=E + N + S + W)
        Button(but, text='Cancel', command=self.onCancel).pack(side=RIGHT)
        Button(but, text='Ok', command=self.onSubmit).pack(side=RIGHT)

    def cmdRad(self):
        self.selectmode = self.varRad.get()

    def onCancel(self):
        self.root.destroy()

    def onSubmit(self):

        t1 = float(self.t1Ent.get())

        # selection

        c1 = less(self.master.X.dis, 1024. + t1)
        c2 = greater_equal(self.master.X.dis, 1024.)

        c = c1 * c2

        self.master.X = self.master.X.selectc(c)

        self.master.display()
        self.master.display_info(self.master.X)

        self.root.destroy()


Dissip(self)



###########################################################################################
#  package:   Mockimgs
#  file:      dzone.py
#  brief:     dzone class
#  copyright: GPLv3
#             Copyright (C) 2023 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

import numpy as np
from astropy import units as u


class Dzone():
  '''
  Define the Dzone class.
  A Dzone is a zone on a detector
  '''
  def __init__(self,detector,xs,ys):
    """
    xs,ys a list of points with unit length
    
    focal    : the telescope focal, this allow to convert x,y from the focal plane
               to angle
    distance : the distance to an object, this allow to convert x,y from the focal plane
               to distances
    
    """
    
    # close the zone
    self.xs = np.concatenate((xs,[xs[0]]))
    self.ys = np.concatenate((ys,[ys[0]]))
    
    # detector
    self.detector = detector
    
    
  
  def x(self,mode=None,unit=None):
    """
    get x values
    
    mode : the mode [None,angle,phys]
    """
    if mode is None or mode=="detector":
      val = self.xs
      
    elif mode=="angle":
      val =  self.detector.toAngle(self.xs)
      
    elif mode=="phys":
      val =  self.detector.toDistance(self.xs)     
    
    if unit is not None:
      val = val.to(unit)
    
    return val  
  
  
  def y(self,mode=None,unit=None):
    """
    get y values
    
    mode : the mode [None,angle,phys]
    """
    if mode is None or mode=="detector":
      val = self.ys
      
    elif mode=="angle":
      val =  self.detector.toAngle(self.ys)
      
    elif mode=="phys":
      val =  self.detector.toDistance(self.ys)     
    
    if unit is not None:
      val = val.to(unit)
    
    return val  










###########################################################################################
#  package:   Mockimgs
#  file:      ifu.py
#  brief:     ifu class
#  copyright: GPLv3
#             Copyright (C) 2023 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

import numpy as np
from astropy import units as u
from .detector import Detector
from .spoints import sPoints

class IFU(Detector):
  '''
  Define the CCD class.
  '''
  def __init__(self,name,size=None):    
        
    self.name  = name
    self.size  = size

    # define the sensible ccd zone
    self.defineSensibleZones()


  def info(self):
    """
    gives info on the ifu
    """
    print("ifu name              : %s"%self.name) 
    print("ifu size              : %s"%self.size)   
    

  def defineSensibleZones(self):
    """
    set the sensible zone in size units
    
    self.sZone is a list of tupple
    """
    
    unit = self.size.unit
    ref  = self.size.ref
    
    xmin = -self.size[0].value/2.
    xmax = +self.size[0].value/2.
    ymin = -self.size[1].value/2.
    ymax = +self.size[1].value/2.  
    
    xs = np.array([xmin,xmax,xmax,xmin,xmin])*unit
    ys = np.array([ymin,ymin,ymax,ymax,ymin])*unit
    
    pts = sPoints(xs,ys,ref)
    self.sZones = [pts]
    
  
  
      
      
        





###########################################################################################
#  package:   Mockimgs
#  file:      scoords.py
#  brief:     scoords class
#  copyright: GPLv3
#             Copyright (C) 2023 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################


from astropy import units
import numpy as np


SKY = "sky"
FOC = "focal plane"
UNI = "universe"


def checkUnit(x,unit,msg=""):
  """
  check the validity of the focal
  """
  if x is None:
    raise ValueError("The value x cannot be None.\n%s"%msg) 
  
  if type(x) is not units.quantity.Quantity:
    raise TypeError("The value x must be of type astropy.units.quantity.Quantity.\n%s"%msg)
  
  if not x.unit.is_equivalent(unit):
    raise ValueError("The unit of the x value must be %s. Found %s.\n%s"%(unit,x.unit,msg))
  
  return True    



def checkDistance(distance):
  return checkUnit(distance,units.cm,msg="The distance must be defined properly.")

def checkFocal(focal):
  return checkUnit(focal,units.cm,msg="The focal must be defined properly.")


  

class sCoords(units.quantity.Quantity):
  """
  This class which derive from the astropy units.quantity.Quantity class
  allows to define a quantity either on a focal plane, the sky or as physical quantities in the universe.
  Providing a focal plane or a distance it allow for a conversion between these three reference frames.
  
  example:
  
  from astropy import units as u
  import numpy as np
  from pNbody.Mockimgs.scoords import sCoords    

  x = sCoords(np.array([1])*u.arcsec,ref="sky")
  x.change_ref("sky")
  x.change_ref("focal plane",focal=1000*u.cm)
  x.change_ref("universe",distance=35*u.Mpc)
  
  x = sCoords(np.array([1])*u.cm,ref="focal plane")
  x.change_ref("universe",distance=35*u.Mpc,focal=1000*u.cm)
  
  """

  def __new__(cls,value,unit=None,dtype=np.inexact,copy=True,order=None,subok=False,ndmin=0,ref=None):  
    new = super().__new__(cls,value,unit=None,dtype=np.inexact,copy=True,order=None,subok=False,ndmin=0)
    # add a reference
    new.ref = ref
    return new

  def to_string(self, unit=None, precision=None, format=None, subfmt=None):
      """
      Generate a string representation of the quantity and its unit.

      The behavior of this function can be altered via the
      `numpy.set_printoptions` function and its various keywords.  The
      exception to this is the ``threshold`` keyword, which is controlled via
      the ``[units.quantity]`` configuration item ``latex_array_threshold``.
      This is treated separately because the numpy default of 1000 is too big
      for most browsers to handle.

      Parameters
      ----------
      unit : unit-like, optional
          Specifies the unit.  If not provided,
          the unit used to initialize the quantity will be used.

      precision : number, optional
          The level of decimal precision. If `None`, or not provided,
          it will be determined from NumPy print options.

      format : str, optional
          The format of the result. If not provided, an unadorned
          string is returned. Supported values are:

          - 'latex': Return a LaTeX-formatted string

          - 'latex_inline': Return a LaTeX-formatted string that uses
            negative exponents instead of fractions

      subfmt : str, optional
          Subformat of the result. For the moment, only used for
          ``format='latex'`` and ``format='latex_inline'``. Supported
          values are:

          - 'inline': Use ``$ ... $`` as delimiters.

          - 'display': Use ``$\\displaystyle ... $`` as delimiters.

      Returns
      -------
      str
          A string with the contents of this Quantity
      """
      
      refstr = " on "+str(self.ref)
      
      if unit is not None and unit != self.unit:
          return self.to(unit).to_string(
              unit=None, precision=precision, format=format, subfmt=subfmt
          )

      formats = {
          None: None,
          "latex": {
              None: ("$", "$"),
              "inline": ("$", "$"),
              "display": (r"$\displaystyle ", r"$"),
          },
      }
      formats["latex_inline"] = formats["latex"]
      
      if format not in formats:
          raise ValueError(f"Unknown format '{format}'")
      elif format is None:
          if precision is None:
              # Use default formatting settings
              return f"{self.value}{self._unitstr:s}{refstr}"
          else:
              # np.array2string properly formats arrays as well as scalars
              return (
                  np.array2string(self.value, precision=precision, floatmode="fixed")
                  + self._unitstr + refstr
              )

       
      # else, for the moment we assume format="latex" or "latex_inline".

      # Set the precision if set, otherwise use numpy default
      pops = np.get_printoptions()
      format_spec = f".{precision if precision is not None else pops['precision']}g"

      def float_formatter(value):
          return Latex.format_exponential_notation(value, format_spec=format_spec)

      def complex_formatter(value):
          return "({}{}i)".format(
              Latex.format_exponential_notation(value.real, format_spec=format_spec),
              Latex.format_exponential_notation(
                  value.imag, format_spec="+" + format_spec
              ),
          )
      
      # The view is needed for the scalar case - self.value might be float.
      latex_value = np.array2string(
          self.view(np.ndarray),
          threshold=(
              conf.latex_array_threshold
              if conf.latex_array_threshold > -1
              else pops["threshold"]
          ),
          formatter={
              "float_kind": float_formatter,
              "complex_kind": complex_formatter,
          },
          max_line_width=np.inf,
          separator=",~",
      )

      latex_value = latex_value.replace("...", r"\dots")

      # Format unit
      # [1:-1] strips the '$' on either side needed for math mode
      if self.unit is None:
          latex_unit = _UNIT_NOT_INITIALISED
      elif format == "latex":
          latex_unit = self.unit._repr_latex_()[1:-1]  # note this is unicode
      elif format == "latex_inline":
          latex_unit = self.unit.to_string(format="latex_inline")[1:-1]

      delimiter_left, delimiter_right = formats[format][subfmt]


      return rf"{delimiter_left}{latex_value} \; {latex_unit}{delimiter_right}"


  def __str__(self):
    return self.to_string()

  def __repr__(self):
      prefixstr = "<" + self.__class__.__name__ + " "
      arrstr = np.array2string(self.view(np.ndarray), separator=", ", prefix=prefixstr)
      refstr = "on "+str(self.ref)
      return f"{prefixstr}{arrstr}{self._unitstr:s} {refstr}>"


  # Arithmetic operations
  #def __mul__(self, other):
  #  """Multiplication between `Quantity` objects and other objects."""
  #  if isinstance(other, (UnitBase, str)):
  #    try:
  #      return self._new_view(self.value.copy(), other * self.unit, propagate_info=False)
  #    except UnitsError:  # let other try to deal with it
  #      return NotImplemented
  #  
  #  return super().__mul__(other)





  def to(self, unit, equivalencies=[], copy=True):
    new = super().to(unit, equivalencies, copy)
    # add a reference
    new.ref = self.ref
    return new


  def change_ref(self,ref,focal=None,distance=None):
    """
    operate some conversions allowing to change the reference system
    """
    
    if self.ref==FOC:

      if   ref==FOC:
        return sCoords(self.copy(),ref=FOC)
              
      elif ref==SKY:
        checkFocal(focal)
        return sCoords(np.arctan(self.copy()/focal),ref=SKY)
        
      elif ref==UNI:
        checkDistance(distance)
        checkFocal(focal)      
        return sCoords(self.copy()/focal*distance,ref=UNI) 
            
    
    elif self.ref==SKY:
      
      if   ref==FOC:
        checkFocal(focal)
        return sCoords(focal*np.tan(self.copy()),ref=FOC)        
      
      elif ref==SKY:
        return sCoords(self.copy(),ref=SKY)
        
      elif ref==UNI:
        checkDistance(distance)
        return sCoords(distance*np.tan(self.copy()),ref=UNI)
      
    elif self.ref==UNI:
      
      if   ref==FOC:
        checkDistance(distance)
        checkFocal(focal)       
        return sCoords(self.copy()/distance*focal,ref=FOC)
      
      elif ref==SKY:
        checkDistance(distance)
        return sCoords(np.arctan(self.copy()/distance),ref=SKY)
        
      elif ref==UNI:
        return sCoords(self.copy(),ref=UNI)      
  
      



###########################################################################################
#  package:   Mockimgs
#  file:      noise.py
#  brief:     noise functions
#  copyright: GPLv3
#             Copyright (C) 2023 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

import numpy as np

def get_std_for_SB_limit(SBlimit,p,area=100,zp=0,sn=3):
  """
  Compute the standard deviation of a Gaussian distribution (noise) 
  that must be applied on pixels to get a given SB limit computed
  in a given area.
  
  SBlimit  : desired SB limit
  p        : pixel sizein arsecond
  area     : area in arcsecond^2
  zp       : zero point
  sn       : signal to noise
  """
  
  # compute sigma
  sigma = p * (np.sqrt(area)/sn) * 10**(-(SBlimit-zp)/2.5)
  
  return sigma








###########################################################################################
#  package:   Mockimags
#  file:      object.py
#  brief:     telescope class
#  copyright: GPLv3
#             Copyright (C) 2023 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

from astropy import units as u
from .los import LineOfSights
from .spoints import sPoints
from scipy.spatial.transform import Rotation
from pNbody import *
import pickle

class Object(Nbody):
  """
  Define an object "object". the latter derives from the Nbody class
  """
  
  def __init__(self,name=None,filename=None,unitsfile=None,nb=None,adjust_coords='remove_identical',
          distance=1*u.Mpc,unit=u.kpc,ref="universe",los=LineOfSights(),rsp_opts={"mode":None,"fac":1},
          ftype='gh5',verbose=0,local=False,status='old'):

    # init an empty object
    super().__init__(ftype=ftype,verbose=verbose,local=local,status=status)

    # additional parameters
    self.name      = name
    self.filename  = filename
    self.distance  = distance
    self.unitsfile = unitsfile
    self.verbose   = verbose
    self.ftype     = ftype

    self.unit     = unit
    self.ref      = ref

    #if adjust_coords is not None:
    #    # APC: optionally remove particles with identical radii. This was
    #    # APC: a hack to work around limitiations of the neighbor-finding
    #    # APC: tree code.
    #    self._remove_identical_coords = 'remove_identical' in adjust_coords
    #    if self._remove_identical_coords:
    #        self.message('WARNING: will remove particles with identical 3D radii')

    # line of sights
    self.los      = los
    # rsp (smoothing options)    
    self.rsp_opts = rsp_opts
        
    # line of sight idx
    self.los_idx  = 0
    
    
    # link towards an instrument
    self.instrument = None
    
    # test nb
    if nb is not None:
      raise ValueError("The option 'nb' is no longer supported.")

  
  def open(self,getAgeMH=False):
    '''  
    Open the N-body model and eventually compute missing quantities. 
    '''  

    if self.filename is not None:
      msg = 'Reading input file(s): {:s}'.format(str(self.filename))
      message.message('Reading input file(s): {:s}'.format(str(self.filename)))
    else:
      return None  
    
    if self.unitsfile is not None:
      message.message('Reading units file: {:s}'.format(str(self.unitsfile)))

    # now open the file
    super().__init__(p_name=self.filename,unitsfile=self.unitsfile,ftype=self.ftype,verbose=self.verbose)
    
    # Remove particles with identical 3D radii
    # APC: This is a hack to work around a limitation of the tree
    # APC: code, which can't cope with identical coordiantes.
    #if self._remove_identical_coords:
    #    self.warning('WARNING: excluding particles with identical 3D radii')
    #    n_before = len(self.rxyz())
    #    u,idx = np.unique(self.rxyz(),return_index=True)
    #    n_after = len(u)
    #    self = self.selectp(lst=self.num[idx])
    #    self.warning('WARNING: removed {:d} of {:d} particles with identical 3D radii'.format(n_before-n_after,n_before))

    
    if not self.has_array("rsp"):
      self.ComputeRsp(5)

    if getAgeMH:

      if not self.has_array("age"):
        # compute Age [in Gyr]
        self.message("Compute ages...")
        self.age = self.StellarAge(units="Gyr")
          
      # compute the metallicity
      if not self.has_array("mh"):   
        self.mh = self.MetalsH()
    

  def info(self):
    """
    give info on the telescope
    """  
    print("object name        : %s"%self.name)
    print("object filename    : %s"%self.filename)
    print("object distance    : %s"%self.distance)
    print("object los         : %s"%self.los.list())


  def get_parameters(self,fmt=None):
    """
    return a dictionary containing usefull parameters
    """
    params = {}

    if fmt=="fits":
      params["OBJ_NAME"]           = (self.name,"Object name")
    else:
      params["object_name"]        = self.name
    
    if fmt=="fits":
      params["OBJ_DIST"]           = (self.distance.to(u.Mpc).value,"Object distance in Mpc")
    else:
      params["object_distance"]    = self.distance

    if fmt=="fits":
      params["OBJ_NLOS"]           = (self.los.n(),"Number of line of sights used")
    else:    
      params["object_nlos"]        = self.los.n()
      
    if fmt=="fits":
      los = self.los.get()
      params["OBJ_LOS"]            = ("[%5.3f,%5.3f,%5.3f]"%(los[0],los[1],los[2]),"line of sight used")
    else:    
      params["object_nlos"]        = self.los.get()     
    
    
    if fmt=="fits":
      params["RSPMODE"]            = (str(self.rsp_opts["mode"]),"RSP mode")
      if self.rsp_opts["mode"]=="const":  
        params["RSPVAL"]           = (self.rsp_opts["val"],"RSP value")
      elif self.rsp_opts["mode"]=="arctan":
        params["RSPMAX"]           = (self.rsp_opts["max"],"RSP max")
        params["RSPSCA"]           = (self.rsp_opts["sca"],"RSP scale")
      elif self.rsp_opts["mode"]=="ln":
        params["RSPSCA"]           = (self.rsp_opts["sca"],"RSP scale")
      else:
        params["RSPFAC"]           = (self.rsp_opts["fac"],"RSP factor")

    else:  
      params["rsp_opts"]           = self.rsp_opts
    
    
    # record time
    
    try:
      tnow = self.Time(units="Gyr")
    except:  
      tnow = -1
      
    if fmt=="fits":  
      params["OBJ_TNOW"]            = (str(tnow),"Simulation Time [Gyr]")
    else:  
      params["obj_tnow "]           = tnow      
    
    
    return params



  def set_los(self,los):
    """
    set line of sights
    """
    self.los = los 

  def set_rsp_opts(self,rsp_opts):
    """
    set rsp options
    """
    self.rsp_opts = rsp_opts



  def getDistance(self):
    """
    return the object distance
    """
    return self.distance
  

  def getFocal(self):
    """
    if defined, get the instrument focal
    """
    if self.instrument is not None:
      return self.instrument.getFocal()
    else:
      raise ValueError("no instrument defined. Cannot get the telescope focal.")
    
    

  def ScaleSmoothingLength(self,x,opt):
    """
    Scale the smoothing length.
    opt : a dictionary
    """
    
    if opt["mode"]=="const":  
      #print("smoothing const: %g"%opt["val"])
      rsp =  np.ones(len(x))* opt["val"]
    
    elif opt["mode"]=="arctan":
      #print("smoothing arctan : %g %g"%(opt["max"],opt["sca"]))
      xmax  = opt["max"]
      scale = opt["sca"]
      rsp =  xmax/(np.pi/2) *np.arctan(x/scale)
  
    elif opt["mode"]=="ln":
      rsp = opt["sca"] * np.log(x/opt["sca"]+1)
    
    else:
      rsp = x
     
    # finally scale with respect to the frsp factor
    rsp = rsp *  opt["fac"]
    
    return rsp
  

  def align(self,pos,axis,axis_ref=[0,0,1]):
    """
    rotate the model with a rotation that align axis with axis_ref
    
    pos      : a set of positions
    axis     : the axis
    axis_ref : the reference axis
    """
    axis1 = axis_ref
    axis2 = axis

    a1 = np.array(axis1, float)
    a2 = np.array(axis2, float)

    a3 = np.array([0, 0, 0], float)
    a3[0] = a1[1] * a2[2] - a1[2] * a2[1]
    a3[1] = a1[2] * a2[0] - a1[0] * a2[2]
    a3[2] = a1[0] * a2[1] - a1[1] * a2[0]

    n1 = np.sqrt(a1[0]**2 + a1[1]**2 + a1[2]**2)
    n2 = np.sqrt(a2[0]**2 + a2[1]**2 + a2[2]**2)
    angle = -np.arccos(np.inner(a1, a2) / (n1 * n2))

    if angle==0:
      return pos
      
    # set the norm of a3 to be angle
    norm = np.sqrt(a3[0]**2 + a3[1]**2 + a3[2]**2)
    a3 = a3/norm*angle  
      
    # create the rotation matix
    R = Rotation.from_rotvec(a3)
    
    # do the rotation
    pos = R.apply(pos)
    
    return pos



    
  def set_for_exposure(self):
    '''
    Set the object for exposure.
    Rotate according to a given los.
    Scale according to a given distance.
    Initial values of the model are modified.
    '''
        
    
    ###################################
    # pos
    
    if self.has_array("pos"):
    
      # rotate
      self.pos = self.align(self.pos,axis=self.los.get())  
      
      # transform in appropriate units
      self.pos  = self.Pos(units="kpc")
        
      x = self.pos[:,0] * u.kpc
      y = self.pos[:,1] * u.kpc 
      z = self.pos[:,2] * u.kpc
          
      # scale the model : kpc -> arcsec
      x = np.arctan(x/self.distance).to(u.arcsec).value
      y = np.arctan(y/self.distance).to(u.arcsec).value
      z = np.arctan(z/self.distance).to(u.arcsec).value
      
      # move back to pos
      self.pos = np.transpose((x,y,z))
      self.pos = self.pos.astype(np.float32)
  
  
    ###################################
    # rsp
    
    if self.has_array("rsp"):
  
      # scale the smoothing length
      self.rsp = self.ScaleSmoothingLength(self.rsp,self.rsp_opts)
    
      # scale : kpc -> arcsec
      self.rsp = self.rsp * u.kpc
      self.rsp = np.arctan(self.rsp/self.distance).to(u.arcsec).value
  
  
    ###################################
    # mass  
  
    if self.has_array("mass"):
      # set proper units
      self.mass = self.Mass(units="Msol")


  
  
  def x(self,mode=None,unit=None):
    """
    get x coordinates of the object
    we assume that x is in kpc
    """
    if mode is None or mode=="phys":
      val = self.pos[:,0]*u.kpc
      
    elif mode=="angle":
      val = np.arctan(self.x()/self.getDistance())
        
    elif mode=="detector":
      val = self.x()/self.getDistance()*self.getFocal()
    
    if unit is not None:
      val = val.to(unit)
    
    return val
      

  def y(self,mode=None,unit=None):
    """
    get y coordinates of the object
    we assume that y is in kpc
    """
    if mode is None or mode=="phys":
      val = self.pos[:,1]*u.kpc
      
    elif mode=="angle":
      val = np.arctan(self.y()/self.getDistance())
        
    elif mode=="detector":
      val = self.y()/self.getDistance()*self.getFocal()
    
    if unit is not None:
      val = val.to(unit)
    
    return val
      
  
  def getPoints(self):
    
    # assume kpc here
    xs = self.pos[:,0]*self.unit
    ys = self.pos[:,1]*self.unit
    
    pts = sPoints(xs,ys,ref=self.ref)
    
    return pts    
        
        

  def draw(self,ax,mode=None,unit=None):
    """
    draw in matplotlib
    do a simple scatter plot
    """

    focal    = self.getFocal()
    distance = self.getDistance()
    
    pts = self.getPoints()
    xs = pts.x(mode=mode,unit=unit,focal=focal,distance=distance)
    ys = pts.y(mode=mode,unit=unit,focal=focal,distance=distance)
      
    ax.scatter(xs,ys,s=2)

    









class oldObject():
  '''
  Define the object class.
  '''
  
  def __init__(self,name=None,filename=None,unitsfile=None,nb=None,adjust_coords='remove_identical',
          distance=1*u.Mpc,unit=u.kpc,ref="universe",los=LineOfSights(),rsp_opts={"mode":None,"fac":1}):
    """
    adjust_coords: string or iterable checked for the following values
        (can be combined):

        'remove' : remove particles with identical 3D radii;
    """
    self.name      = name
    self.filename  = filename
    self.distance  = distance
    self.unitsfile = unitsfile

    self.unit     = unit
    self.ref      = ref

    if adjust_coords is not None:
        # APC: optionally remove particles with identical radii. This was
        # APC: a hack to work around limitiations of the neighbor-finding
        # APC: tree code.
        self._remove_identical_coords = 'remove_identical' in adjust_coords
        if self._remove_identical_coords:
            message.message('WARNING: will remove particles with identical 3D radii',verbosity=1,level=2,color="r")

    # line of sights
    self.los      = los
    # rsp (smoothing options)    
    self.rsp_opts = rsp_opts
        
    # line of sight idx
    self.los_idx  = 0
    
    # if nb is given, make it global
    self.nb = nb
    
    # link towards an instrument
    self.instrument = None
    
    

  def info(self):
    """
    give info on the telescope
    """  
    print("object name        : %s"%self.name)
    print("object filename    : %s"%self.filename)
    print("object distance    : %s"%self.distance)
    print("object los         : %s"%self.los.list())


  def get_parameters(self,fmt=None):
    """
    return a dictionary containing usefull parameters
    """
    params = {}
    
    if fmt=="fits":
      params["OBJ_DIST"]           = (self.distance.to(u.Mpc).value,"Object distance in Mpc")
    else:
      params["object_distance"]    = self.distance

    if fmt=="fits":
      params["OBJ_NLOS"]           = (self.los.n(),"Number of line of sights used")
    else:    
      params["object_nlos"]        = self.los.n()
      
    if fmt=="fits":
      los = self.los.get()
      params["OBJ_LOS"]            = ("[%5.3f,%5.3f,%5.3f]"%(los[0],los[1],los[2]),"line of sight used")
    else:    
      params["object_nlos"]        = self.los.get()     
    
    
    if fmt=="fits":
      params["RSPMODE"]            = (str(self.rsp_opts["mode"]),"RSP mode")
      if self.rsp_opts["mode"]=="const":  
        params["RSPVAL"]           = (self.rsp_opts["val"],"RSP value")
      elif self.rsp_opts["mode"]=="arctan":
        params["RSPMAX"]           = (self.rsp_opts["max"],"RSP max")
        params["RSPSCA"]           = (self.rsp_opts["sca"],"RSP scale")
      elif self.rsp_opts["mode"]=="ln":
        params["RSPSCA"]           = (self.rsp_opts["sca"],"RSP scale")
      else:
        params["RSPFAC"]           = (self.rsp_opts["fac"],"RSP factor")

    else:  
      params["rsp_opts"]           = self.rsp_opts
    
    
    return params



  def set_los(self,los):
    """
    set line of sights
    """
    self.los = los 

  def set_rsp_opts(self,rsp_opts):
    """
    set rsp options
    """
    self.rsp_opts = rsp_opts

  def open(self):
    '''  
    Open the N-body model and eventually compute missing quantities. 
    '''  
    if self.filename is not None:
      msg = 'Reading input file(s): {:s}'.format(str(self.filename))
      message.message('Reading input file(s): {:s}'.format(str(self.filename)))
    else:
      return None  
    
    if self.unitsfile is not None:
      message.message('Reading units file: {:s}'.format(str(self.unitsfile)))

    # open the object
    nb = Nbody(self.filename,unitsfile=self.unitsfile)

    # Remove particles with identical 3D radii
    # APC: This is a hack to work around a limitation of the tree
    # APC: code, which can't cope with identical coordiantes.
    if self._remove_identical_coords:
        message.message('WARNING: excluding particles with identical 3D radii',verbosity=1,level=2,color="r")
        n_before = len(nb.rxyz())
        u,idx = np.unique(nb.rxyz(),return_index=True)
        n_after = len(u)
        nb = nb.selectp(lst=nb.num[idx])
        message.message('WARNING: removed {:d} of {:d} particles with identical 3D radii'.format(n_before-n_after,n_before),verbosity=1,level=2,color="r")

    if not nb.has_array("rsp"):
      nb.ComputeRsp(5)

    if not nb.has_array("age"):
      # compute Age [in Gyr]
      print("Compute ages...")
      nb.age = nb.StellarAge(units="Gyr")
      print("done.")
  
  
    # compute the metallicity
    if not nb.has_array("mh"):   
      nb.mh = nb.MetalsH()
    
    # make global
    self.nb = nb  
  

  def getDistance(self):
    """
    return the object distance
    """
    return self.distance
  

  def getFocal(self):
    """
    if defined, get the instrument focal
    """
    if self.instrument is not None:
      return self.instrument.getFocal()
    else:
      raise ValueError("no instrument defined. Cannot get the telescope focal.")
    
    

  def ScaleSmoothingLength(self,x,opt):
    """
    Scale the smoothing length.
    opt : a dictionary
    """
    
    if opt["mode"]=="const":  
      #print("smoothing const: %g"%opt["val"])
      rsp =  np.ones(len(x))* opt["val"]
    
    elif opt["mode"]=="arctan":
      #print("smoothing arctan : %g %g"%(opt["max"],opt["sca"]))
      xmax  = opt["max"]
      scale = opt["sca"]
      rsp =  xmax/(np.pi/2) *np.arctan(x/scale)
  
    elif opt["mode"]=="ln":
      rsp = opt["sca"] * np.log(x/opt["sca"]+1)
    
    else:
      rsp = x
     
    # finally scale with respect to the frsp factor
    rsp = rsp *  opt["fac"]
    
    return rsp
  

  def align(self,pos,axis,axis_ref=[0,0,1]):
    """
    rotate the model with a rotation that align axis with axis_ref
    
    pos      : a set of positions
    axis     : the axis
    axis_ref : the reference axis
    """
    axis1 = axis_ref
    axis2 = axis

    a1 = np.array(axis1, float)
    a2 = np.array(axis2, float)

    a3 = np.array([0, 0, 0], float)
    a3[0] = a1[1] * a2[2] - a1[2] * a2[1]
    a3[1] = a1[2] * a2[0] - a1[0] * a2[2]
    a3[2] = a1[0] * a2[1] - a1[1] * a2[0]

    n1 = np.sqrt(a1[0]**2 + a1[1]**2 + a1[2]**2)
    n2 = np.sqrt(a2[0]**2 + a2[1]**2 + a2[2]**2)
    angle = -np.arccos(np.inner(a1, a2) / (n1 * n2))

    if angle==0:
      return pos
      
    # set the norm of a3 to be angle
    norm = np.sqrt(a3[0]**2 + a3[1]**2 + a3[2]**2)
    a3 = a3/norm*angle  
      
    # create the rotation matix
    R = Rotation.from_rotvec(a3)
    
    # do the rotation
    pos = R.apply(pos)
    
    return pos



    
  def set_for_exposure(self):
    '''
    Set the object for exposure.
    Rotate according to a given los.
    Scale according to a given distance.
    Initial values of the model are modified.
    '''
    
    nb = self.nb
    
    
    ###################################
    # pos
    
    if nb.has_array("pos"):
    
      # rotate
      nb.pos = self.align(nb.pos,axis=self.los.get())  
      
      # transform in appropriate units
      nb.pos  = nb.Pos(units="kpc")
        
      x = nb.pos[:,0] * u.kpc
      y = nb.pos[:,1] * u.kpc 
      z = nb.pos[:,2] * u.kpc
          
      # scale the model : kpc -> arcsec
      x = np.arctan(x/self.distance).to(u.arcsec).value
      y = np.arctan(y/self.distance).to(u.arcsec).value
      z = np.arctan(z/self.distance).to(u.arcsec).value
      
      # move back to pos
      nb.pos = np.transpose((x,y,z))
      nb.pos = nb.pos.astype(np.float32)
  
  
    ###################################
    # rsp
    
    if nb.has_array("rsp"):
  
      # scale the smoothing length
      nb.rsp = self.ScaleSmoothingLength(nb.rsp,self.rsp_opts)
    
      # scale : kpc -> arcsec
      nb.rsp = nb.rsp * u.kpc
      nb.rsp = np.arctan(nb.rsp/self.distance).to(u.arcsec).value
  
  
    ###################################
    # mass  
  
    if nb.has_array("mass"):
      # set proper units
      nb.mass = nb.Mass(units="Msol")


  
    # make global
    self.nb = nb
  
  
  def x(self,mode=None,unit=None):
    """
    get x coordinates of the object
    we assume that x is in kpc
    """
    if mode is None or mode=="phys":
      val = self.nb.x()*u.kpc
      
    elif mode=="angle":
      val = np.arctan(self.x()/self.getDistance())
        
    elif mode=="detector":
      val = self.x()/self.getDistance()*self.getFocal()
    
    if unit is not None:
      val = val.to(unit)
    
    return val
      

  def y(self,mode=None,unit=None):
    """
    get y coordinates of the object
    we assume that y is in kpc
    """
    if mode is None or mode=="phys":
      val = self.nb.y()*u.kpc
      
    elif mode=="angle":
      val = np.arctan(self.y()/self.getDistance())
        
    elif mode=="detector":
      val = self.y()/self.getDistance()*self.getFocal()
    
    if unit is not None:
      val = val.to(unit)
    
    return val
      
  
  def getPoints(self):
    
    # assume kpc here
    xs = self.nb.x()*self.unit
    ys = self.nb.y()*self.unit
    
    pts = sPoints(xs,ys,ref=self.ref)
    
    return pts    
        
        

  def draw(self,ax,mode=None,unit=None):
    """
    draw in matplotlib
    do a simple scatter plot
    """

    focal    = self.getFocal()
    distance = self.getDistance()
    
    pts = self.getPoints()
    xs = pts.x(mode=mode,unit=unit,focal=focal,distance=distance)
    ys = pts.y(mode=mode,unit=unit,focal=focal,distance=distance)
      
    ax.scatter(xs,ys,s=2)



#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      homosphere.py
#  brief:     Defines homogeneous sphere profiles
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################


"""
homogeneous mass model
"""

import numpy as np

def getRho(M, R):
    """
    return the sphere density
    """
    return M/(4/3.*np.pi * R**3)


def Potential(M, R, r,G=1.):
    """
    Potential
    """
    rho = getRho(M, R)
    
    def pot_int(r):
      return 2/3.*np.pi * G * rho * r**2  - 2*np.pi * G * rho * R**2
    
    def pot_ext(r):
      return -G*M/r
    
    
    if type(r) == np.ndarray:
      return np.where(r<R, pot_int(r), pot_ext(r)) 
    
    else:
      if r<R:
        return pot_int(r)
      else:
        return pot_ext(r)
        
        
def Density(M, R, r,G=1.):
    """
    Density
    """
    rho = getRho(M, R)
    
    def dens_int(r):
      return rho
    
    def dens_ext(r):
      return 0
    
    
    if type(r) == np.ndarray:
      return np.where(r<R, dens_int(r), dens_ext(r)) 
    
    else:
      if r<R:
        return dens_int(r)
      else:
        return dens_ext(r)    
    
    


def Vcirc(M, R, r, G=1.):
    """
    Circular Velocity
    """
    rho = getRho(M, R)
    

    def vc_int(r):
      return np.sqrt(4/3.*np.pi*G*rho)*r
    
    def vc_ext(r):
      return np.sqrt(4/3.*np.pi*G*rho*R**3/r)
    
    
    if type(r) == np.ndarray:
      return np.where(r<R, vc_int(r), vc_ext(r)) 
    
    else:
      if r<R:
        return vc_int(r)
      else:
        return vc_ext(r)   






class Nbody:
  
  def print_transformations(self):
    print("I am transformations")

  def UU(self):
    """
    Return the gas specific energy of the model.
    The output is an nx1 float array.
    """
    return self.u




#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      MIST.py
#  brief:     
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

import numpy as np
import tarfile
import glob
import os
from pNbody.Isochrones.main import  IsochronesGrid, IsochroneBlock


def read_mesa_iso(filename,MBs=[],default_keys=None):
  """
  read one mesa iso file
  and return it content as an array
  
  filename      : name of the file to read
  default_keys  : default keys to read. If None, ignore
  
  """

  def read_block(f):
    
    # read block header
    f.readline()
    line = f.readline()
    line = line.split("=")[1]
    line = line.split()
    nrows = int(line[0])
    ncols = int(line[1])
    f.readline()
    
    # keys
    keys = f.readline()
    keys = keys.split()[1:]
    # make a dict
    keys = {x: i for i,x in enumerate(keys)}
    
    if default_keys is not None:
      new_keys = {}
      tmp_keys = []
      for i,k in enumerate(default_keys):
        new_keys[k] = i
        tmp_keys.append(keys[k])
      keys = new_keys  
    else:
      tmp_keys = []
      for i,k in enumerate(keys):
        tmp_keys.append(keys[k])
   
    # read 
    data = np.loadtxt(f,max_rows=nrows,usecols=tmp_keys)
    
    # skip the last line
    f.readline()
    
    return data,keys
  
  def toMistBlock(data,keys):
    "convert an array to a MistBlock object"
    
    
    MB = IsochroneBlock(data=data,keys=keys)
    
    # min/max stellar mass in solar masses
    MB.Mmin = min(MB.get("initial_mass"))
    MB.Mmax = max(MB.get("initial_mass"))
    MB.MH   =     MB.get("[Fe/H]_init")[0]
    # age in Myr
    MB.Age  =     10**MB.get("log10_isochrone_age_yr")[0] / 1e6
    
    return MB
  

  
  print("reading: ",filename)
  f = open(filename,'r')
  f.seek(0,2)
  size = f.tell()
  f.seek(0,0)
    
  # skip file header
  for i in range(9):
    f.readline()
  
  # loop over all blocks
  nblocks = 0
  while True:
    try:
      # read the block
      data,keys = read_block(f)
            
      # convert to a block object
      MB = toMistBlock(data,keys)
      MBs.append(MB)
      
      nblocks = nblocks + 1
    except:
      break
  
  return MBs



###########################################################
# 
# Isochrone Grid
#
###########################################################

class Grid(IsochronesGrid):
  
  def Init(self):

    # get all files
    self.filenames = glob.glob(os.path.join(self.dirname,"*"))

    # list of Isocrhones blocks
    self.MBs = []    
        
    # read all the tar file and fill the isochrones list
    self.readAll() 
    
    # create the DB (self.M) from the isochrones list
    self.organizeDB()


  def readAll(self):
    "read the data base by looping over files"
    
    for filename in self.filenames:  
      self.MBs = read_mesa_iso(filename,MBs=self.MBs,default_keys=self.default_keys)

    # define keys
    self.keys = self.MBs[0].keys










#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      hernquist.py
#  brief:     Defines hernquist profiles
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################


"""
hernquist model
"""

import numpy as np

def Potential(rho0, a, r, G=1.):
    """
    Potential
    """
    return -4*np.pi*G*rho0*a**2 * 1/(2*(1+(r/a)))



def CumulativeMass(rho0, a, r, G=1.):
    """
    Hernquist cumulative mass
    """
    return   4*np.pi*rho0*a**3 *   0.5*(r/a)**2/(1+(r/a))**2

def TotalMass(rho0, a, G=1.):
    """
    Hernquist total mass
    """
    return   4*np.pi*rho0*a**3 *   0.5
    

def Density(rho0, a, r, G=1.):
    """
    Density
    """
    return rho0 / ( (r/a)**1 * (1+(r/a))**3  )
    

def Vcirc(rho0, a, r, G=1.):
    """
    circular velocity
    """
        
    Mr = 4*np.pi*rho0*a**3 * (  (r/a)**2 / (2*(1+r/a)**2)  )
    return np.sqrt(G*Mr/r)
    




#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      shell.py
#  brief:     Defines shell profiles
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################


"""
ring model

Lass & Blitzer 1982
see also galpy RingPotential.py (Bovy)

"""

import numpy as np
from scipy import special


def Potential(M, R0, R, z, G=1.):
    """
    Shell Potential
    """
    
    m= 4.*R*R0/((R+R0)**2+z**2)
    P = - 2*G*M/np.pi/np.sqrt( (R0 + R)**2 + z**2 ) *special.ellipk(m)
  
    return P
    
    
    
def dPotential(M, R0, R, z, G=1.):
    """
    Shell first derivative of Potential
    """
    
    m= 4.*R*R0/((R+R0)**2+z**2)
    dP = G*M/np.pi/R/np.sqrt( (R0 + R)**2 + z**2 ) * ( m* (R**2-R0**2-z**2)/(4*(1-m)*R*R0)* special.ellipe(m) + special.ellipk(m) )
    
    return dP
    
    
    
    
        
    
    
    
    
    



#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      isochrone.py
#  brief:     Defines isochrone profiles
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################


"""
isochrone model
"""

import numpy as np


def Potential(M, b, r, G=1.):
    """
    Potential
    """
    return -G * M / (b + np.sqrt(r**2 + b**2))


def Density(M, b, r, G=1.):
    """
    Density
    """
    a = np.sqrt(b**2+r**2)
    return M* (3*(b+a)*a**2 - r**2*(b+3*a))/( 4*np.pi * (b+a)**3 * a**3 )
    


def Vcirc(M, b, r, G=1.):
    """
    circular velocity
    """
    a = np.sqrt(b**2+r**2)
    return np.sqrt(G * M * r**2 / ( a*(b+a)**2) )
